# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from typing import Callable, Optional

import numpy as np
from owlready2 import Thing
from pyiron_atomistics import Project
from pyiron_base import GenericJob
from ryvencore import Node as NodeCore
from ryvencore.Base import Event
from ryvencore.InfoMsgs import InfoMsgs
from ryvencore.NodePort import NodePort
from ryvencore.utils import deserialize

from ironflow.gui.workflows.canvas_widgets.nodes import NodeWidget
from ironflow.model import dtypes
from ironflow.model.otypes import otype_from_str
from ironflow.model.port import NodeInput, NodeInputBP, NodeOutput, NodeOutputBP
from ironflow.utils import display_string


class PortList(list):
    """
    When used to hold a collection of `NodePort` objects, the values of these ports then become accessible by their
    labels, as long as those labels do not match an existing method of the builtin list class.

    Warning:
        This class makes no check that these labels are unique; if multiple items have the same label, the first one
        is returned.

    Warning:
        Accessing port values in this way side-steps ryven functionality when in exec mode or using an executor
        (i.e. when `running_with_executor`).
    """

    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)
        self._port_finder = PortFinder(self)
        self._value_finder = ValueFinder(self)
        # This additional mis-direction is necessary so that ports can have the same labels as list class methods

    @property
    def ports(self):
        """
        Allows attribute-like access to ports by their `label_str`
        """
        return self._port_finder

    @property
    def values(self):
        """
        Allows attribute-like access to port values by their `label_str`

        Calling `port_list.values.some_label` is equivalent to `port_list.ports.some_label.val`
        """
        return self._value_finder

    @property
    def labels(self):
        return [item.label_str if isinstance(item, NodePort) else None for item in self]


class PortFinder:
    def __init__(self, port_list: PortList):
        self._port_list = port_list

    @property
    def _filtered_port_list(self):
        return [item for item in self._port_list if isinstance(item, NodePort)]

    def __getattr__(self, key):
        for node_port in self._filtered_port_list:
            if node_port.label_str == key:
                return node_port
        raise AttributeError(f"No port found with the label {key}")

    def __getitem__(self, item):
        return self.__getattr__(item)

    def __iter__(self):
        return self._filtered_port_list.__iter__()

    def __len__(self):
        return len(self._port_list)


class ValueFinder(PortFinder):
    def __getattr__(self, key):
        node_port = super().__getattr__(key)
        return node_port.val

    def __iter__(self):
        val_list = [p.val for p in self._filtered_port_list]
        return val_list.__iter__()


class Node(NodeCore):
    """
    A parent class for all ironflow nodes. Apart from a small quality-of-life difference where outputs are
    accessible in the same way as inputs (i.e. with a method `output(i)`), the main change here is the `before_update`
    and `after_update` events. Callbacks to happen before and after the update can be added to (removed from) these with
    the `connect` (`disconnect`) methods on the event. Such callbacks need to take the node itself as the first
    argument, and the integer specifying which input channel is being updated as the second argument.

    Also provides a "representation" that gets used in the GUI to give a more detailed look at node data, which defaults
    to showing output channel values.

    Children should specify a title and some combination of initial input, output, and what to do when updated, e.g.:

    >>> class My_Node(Node):
    >>>     title = "MyUserNode"
    >>>     init_inputs = [
    >>>         NodeInputBP(dtype=dtypes.Integer(default=1), label="foo")
    >>>     ]
    >>>     init_outputs = [
    >>>        NodeOutputBP(label="bar")
    >>>     ]
    >>>     color = 'cyan'
    >>>
    >>> def update_event(self, inp=-1):
    >>>     self.set_output_val(0, self.input(0) + 42)

    Note:
        When registering nodes from a module or .py file, only children of this class with names ending in `_Node` will
        get registered.
    """

    main_widget_class = NodeWidget

    color = "#ff69b4"  # Add an abrasive default color -- won't crash if you forget to add one, but pops out a bit

    def __init__(self, params):
        super().__init__(params)
        self.inputs = PortList()
        self.outputs = PortList()

        self.before_update = Event(self, int)
        self.after_update = Event(self, int)
        self.actions = (
            dict()
        )  # Resolves Todo from ryven.NENV, moving it to our node class instead of ryvencore

        self.representation_updated = False
        self.after_update.connect(self._representation_update)

    def _add_io(self, io_group, new_io, insert: int = None):
        if insert is not None:
            io_group.insert(insert, new_io)
        else:
            io_group.append(new_io)

    def create_input(
        self,
        type_: str = "data",
        label: str = "",
        add_data: Optional[dict] = None,
        dtype: Optional[dtypes.DType] = None,
        otype: Optional[Thing] = None,
        insert: Optional[int] = None,
    ):
        """Creates and add a new input port"""
        inp = NodeInput(
            node=self,
            type_=type_,
            label_str=label,
            add_data=add_data,
            dtype=dtype,
            otype=otype,
        )
        self._add_io(self.inputs, inp, insert=insert)

    def create_input_dt(
        self, dtype: dtypes.DType, label: str = "", add_data={}, insert: int = None
    ):
        raise RuntimeError(
            "Ironflow uses custom NodePort classes and this method is not supported. "
            "Please use create_input instead."
        )

    def create_output(
        self,
        type_: str = "data",
        label: str = "",
        dtype: Optional[dtypes.DType] = None,
        otype: Optional[Thing] = None,
        insert: Optional[int] = None,
    ):
        """Create and add a new output port"""
        out = NodeOutput(
            node=self,
            type_=type_,
            label_str=label,
            dtype=dtype,
            otype=otype,
        )
        self._add_io(self.outputs, out, insert=insert)

    def _load_otype(self, data: dict):
        try:
            return otype_from_str(data["otype_namespace"], data["otype_name"])
        except KeyError:
            return None

    def setup_ports(self, inputs_data=None, outputs_data=None):
        # A streamlined version of the ryvencore method which exploits our NodeInput
        # and NodeOutput classes instead, and for which all ports have a dtype
        if not inputs_data and not outputs_data:
            for i in range(len(self.init_inputs)):
                inp = self.init_inputs[i]
                self.create_input(
                    type_=inp.type_,
                    label=inp.label,
                    add_data=inp.add_data,
                    dtype=inp.dtype,
                    otype=inp.otype,
                )

            for o in range(len(self.init_outputs)):
                out = self.init_outputs[o]
                self.create_output(
                    type_=out.type_,
                    label=out.label,
                    dtype=out.dtype,
                    otype=out.otype,
                )

        else:
            # load from data
            # initial ports specifications are irrelevant then

            for inp in inputs_data:
                dtype = dtypes.DType.from_str(inp["dtype"])(
                    _load_state=deserialize(inp["dtype state"])
                )
                self.create_input(
                    type_=inp["type"],
                    label=inp["label"],
                    add_data=inp,
                    dtype=dtype,
                    otype=self._load_otype(inp),
                )

                if "val" in inp:
                    # this means the input is 'data' and did not have any connections,
                    # so we saved its value which was probably represented by some
                    # widget in the front end which has probably overridden the
                    # Node.input() method
                    self.inputs[-1].val = deserialize(inp["val"])
                    self.inputs[-1].set_dtype_ok()

            for out in outputs_data:
                dtype = dtypes.DType.from_str(out["dtype"])(
                    _load_state=deserialize(out["dtype state"])
                )
                self.create_output(
                    type_=out["type"],
                    label=out["label"],
                    dtype=dtype,
                    otype=self._load_otype(out),
                )

    @property
    def all_input_is_valid(self):
        return all(p.ready for p in self.inputs.ports)

    def place_event(self):
        # place_event() is executed *before* the connections are built
        super().place_event()
        for inp in self.inputs:
            if (
                inp.val is None
            ):  # Don't over-write data from loaded sessions with defaults
                if inp.dtype is not None:
                    inp.update(inp.dtype.default)
                elif "val" in inp.add_data.keys():
                    inp.update(inp.add_data["val"])

    def update(self, inp=-1):
        self.before_update.emit(self, inp)
        super().update(inp=inp)
        self.after_update.emit(self, inp)

    def output(self, i):
        return self.outputs[i].val

    @staticmethod
    def _representation_update(self, inp):
        self.representation_updated = True

    @property
    def _source_code(self):
        try:
            return inspect.getsource(self.__class__)
        except TypeError:
            # Classes defined in the notebook can't access their source this way
            return ""

    @property
    def _standard_representations(self):
        standard_reps = {
            o.label_str if o.label_str != "" else f"output{i}": o.val
            for i, o in enumerate(self.outputs)
            if o.type_ == "data"
        }
        try:
            standard_reps["source code"] = display_string(self._source_code)
        except OSError:
            # No source code can be found for user nodes defined in the nodebook
            # But the source is available there anyhow
            pass
        return standard_reps

    @property
    def extra_representations(self):
        """
        When developing nodes, override this with any desired additional representations.

        Note that standard representations exist for all output ports using the port's label (where available), so if
        you add a key here matching one of those labels, you will override the standard output.
        """
        return {}

    @property
    def representations(self) -> dict:
        return {**self._standard_representations, **self.extra_representations}

    def set_all_outputs_to_none(self):
        for i in range(len(self.outputs)):
            self.set_output_val(i, None)


class PlaceholderWidgetsContainer:
    """
    An object that just returns None for all accessed attributes so widgets.MyWidget in the non-ironflow nodes files
    just returns None.
    """

    def __getattr__(self, item):
        return None


class BatchingNode(Node, ABC):
    """
    A node whose update behaviour supports batched inputs.
    """

    @property
    def batched_inputs(self):
        return {
            inp.label_str: inp
            for inp in self.inputs
            if inp.type_ == "data" and inp.dtype.batched
        }

    @property
    def unbatched_inputs(self):
        return {
            inp.label_str: inp
            for inp in self.inputs
            if inp.type_ == "data" and not inp.dtype.batched
        }

    @property
    def batched(self):
        return len(self.batched_inputs) > 0

    @property
    def batch_lengths(self):
        return {k: len(v.val) for k, v in self.batched_inputs.items()}

    @property
    def _unbatched_kwargs(self):
        return {k: v.val for k, v in self.unbatched_inputs.items()}

    @property
    def _batched_kwargs(self):
        return [
            {
                k: v.val[i]
                for k, v in zip(
                    self.batched_inputs.keys(), self.batched_inputs.values()
                )
            }
            for i in range(list(self.batch_lengths.values())[0])
        ]

    def generate_output(self) -> dict:
        if self.batched:
            batch_length_vals = list(self.batch_lengths.values())
            if len(batch_length_vals) > 0 and not np.all(
                np.array(batch_length_vals) == batch_length_vals[0]
            ):
                raise ValueError(
                    f"Not all batch lengths are the same: {self.batch_lengths}"
                )
            return self.generate_batched_output()
        else:
            return self.generate_unbatched_output()

    def generate_unbatched_output(self):
        return self.node_function(**self._unbatched_kwargs)

    def generate_batched_output(self):
        outputs = []
        for i, kwargs in enumerate(self._batched_kwargs):
            kwargs.update(self._unbatched_kwargs)
            outputs.append(self.node_function(batch_index=i, **kwargs))
        return {key: [d[key] for d in outputs] for key in outputs[0].keys()}

    def set_output(self):
        try:
            output = self.generate_output()
        except Exception as e:
            self.clear_output()
            raise e
        for k, v in output.items():
            self.outputs.ports[k].dtype.batched = self.batched
            self.outputs.ports[k].set_val(v)

    def clear_output(self):
        for p in self.outputs.ports:
            if p.type_ == "data":
                p.set_val(None)
                p.dtype.batched = self.batched

    def batched_representation(
        self, label: str, representation_function: Callable, *args
    ) -> dict | None:
        """
        Batched output requires multiple representations instead of a single one. Use
        this function to wrap a function that produces your desired representation.
        Resulting batched representations simply get an index added to their label.

        Args:
            label (str): The name of the representation.
            representation_function (Callable): A function producing the representation.
            *args: Members of `self.inputs.values` or `self.outputs.values` needed for
                the representation.

        Returns:
            (dict): The representation(s).

        Examples:
            >>> def extra_representations(self):
            >>>     return {
            >>>         **self.batched_representation(
            >>>             "bigger", self._add5, self.outputs.values.n
            >>>        )
            >>>     }
            >>>
            >>> @staticmethod
            >>> def _add5(n: int):
            >>>     return n + 5
        """
        try:
            if self.batched:
                return {
                    f"{label}_{i}": representation_function(*batch_args)
                    for i, batch_args in enumerate(zip(*args))
                }
            else:
                return {label: representation_function(*args)}
        except Exception as e:
            return {label: f"Failed with: {e}"}

    @abstractmethod
    def node_function(self, *args, **kwargs) -> dict:
        """
        Takes all data input as kwargs, must return a dict with one entry for each data
        output
        """
        pass


class DataNode(BatchingNode, ABC):
    """
    A node that can update as soon as all input is valid and produces output data.
    """

    def update_event(self, inp=-1):
        if self.all_input_is_valid:
            self.set_output()
        else:
            self.clear_output()

    def place_event(self):
        super().place_event()
        self.update()


class JobNode(BatchingNode, ABC):
    """
    A parent class for nodes that run a pyiron job.

    Child classes are required to specify a `_generate_job` method, which takes the
    node input data and returns a `pyiron_base.GenericJob` object, and (optionally) a
    `_get_output_from_job` method which takes the executed job and the node input data
    and returns a dictionary of output with keys corresponding to node output port
    labels. The `job` output type can be made more specific with the `valid_job_classes`
    class attribute.

    In the event that some of the input is batched but the `name` input is *not*
    batched, then the `name` argument passed to `_generate_job` is automatically
    appended with batch index information to prevent jobs from having the same name.

    The node has `run` and `remove` input exec ports and a `ran` output exec port.
    Once the node has been run, all inputs get locked and remain locked until the
    removal process is triggered. Remove clears all the outputs and sets them to `None`.

    If a failure is encountered during the `run` process, the exception is raised, all
    pyiron jobs are deleted (this might need to change if we want to do more expensive
    jobs in the future!), the node output is cleared, and the offending exception is
    raised to the log.

    All descendant classes from this node class expect to have `run`, `remove`, and
    `name` input, and `ran` output. Therefore, when defining additional ports, use this
    format:

    >>> init_inputs = JobNode.init_inputs + [WHATEVER_ELSE_YOU_WANT]
    >>> init_outputs = JobNode.init_outputs + [OTHER_STUFF]
    """

    color = "#c4473f"
    valid_job_classes = None
    init_inputs = [
        NodeInputBP(type_="exec", label="run"),
        NodeInputBP(type_="exec", label="reset"),
        NodeInputBP(dtype=dtypes.String(default="calc"), label="name"),
    ]
    init_outputs = [
        NodeOutputBP(type_="exec", label="ran"),
        NodeOutputBP(dtype=dtypes.Data(valid_classes=GenericJob), label="job"),
    ]

    def place_event(self):
        super().place_event()
        if self.valid_job_classes is not None:
            self.outputs.ports.job.dtype.valid_classes = self.valid_job_classes

    def update_event(self, inp=-1):
        if inp == 0 and (not self.block_updates) and self.all_input_is_valid:
            self.block_updates = True
            try:
                self._jobs = []
                self.set_output()
                self.exec_output(0)
            except Exception as e:
                self._on_remove_signal()
                raise e
        else:
            self.clear_output()

    def update(self, inp=-1):
        if inp == 1:
            # Bypass the `lock_updates` to delete the executed job and unlock updates
            self._on_remove_signal()
            # self.update(-1)
        else:
            super().update(inp=inp)

    def node_function(self, name, *args, **kwargs):
        if self.batched and not self.inputs.ports.name.dtype.batched:
            name = f"{name}_batch{kwargs['batch_index']}"
        job = self._raise_error_if_not_initialized(
            self._generate_job(name, *args, **kwargs)
        )
        self._jobs.append(job)
        job.run()
        data = self._get_output_from_job(job, *args, name=name, **kwargs)
        return {
            "job": job,
            **data,
        }

    def _on_remove_signal(self):
        if hasattr(self, "_jobs"):
            for job in self._jobs:
                job.remove()
        self.clear_output()
        self.block_updates = False

    def _raise_error_if_not_initialized(self, job: GenericJob) -> GenericJob:
        if job.status == "initialized":
            return job
        else:
            self.block_updates = False
            raise RuntimeError(
                f"The job {self.inputs.values.name} already exists. Delete it first or"
                f"choose a different name."
            )

    @abstractmethod
    def _generate_job(self, name: str, *args, **kwargs) -> GenericJob:
        """
        Takes a (potentially modified[^1]) name and the rest of the input data and
        returns a job to `.run()`

        [^1]: If some of the input is batched, but the `name` field is not batched, the
              batch index information will get automatically appended to the name.
        """

        pass

    def _get_output_from_job(self, finished_job: GenericJob, *args, **kwargs) -> dict:
        """
        Takes the executed job and all the node input data, and should return relevant
        output data as a dictionary with keys matching the output port labels.
        """

        return {}


class JobMaker(JobNode, ABC):
    """
    A job-running node that takes creates a new job instance from scratch, ready to
    `.run()`.

    All descendant classes from this node class expect to have `run`, `remove`, `name`,
    and `project` input, and `ran` output. Therefore, when defining additional ports,
    use this format:

    >>> init_inputs = JobMaker.init_inputs + [WHATEVER_ELSE_YOU_WANT]
    >>> init_outputs = JobMaker.init_outputs + [OTHER_STUFF]
    """

    init_inputs = JobNode.init_inputs + [
        NodeInputBP(dtype=dtypes.Data(valid_classes=Project), label="project")
    ]


class JobTaker(JobNode, ABC):
    """
    A job-running node that takes a template job instance as input, and copies and
    modifies it prior to use.

    Valid classes for the `job` input can be overriden with the `valid_job_classes`
    attribute.

    All descendant classes from this node class expect to have `run`, `remove`, `name`,
    and `job` input, and `ran` output. Therefore, when defining additional ports, use
    this format:

    >>> init_inputs = JobTaker.init_inputs + [WHATEVER_ELSE_YOU_WANT]
    >>> init_outputs = JobTaker.init_outputs + [OTHER_STUFF]
    """

    init_inputs = JobNode.init_inputs + [
        NodeInputBP(dtype=dtypes.Data(valid_classes=GenericJob), label="job")
    ]

    def place_event(self):
        super().place_event()
        if self.valid_job_classes is not None:
            self.inputs.ports.job.dtype.valid_classes = self.valid_job_classes

    def _generate_job(self, name: str, job: GenericJob, **kwargs) -> GenericJob:
        copied_job = job.copy_to(new_job_name=name)
        return self._modify_job(copied_job, **kwargs)

    @abstractmethod
    def _modify_job(self, copied_job: GenericJob, *args, **kwargs) -> GenericJob:
        """
        Takes the generated job, modifies it in place using the input data (except for
        the `job` and `name` input values) and returns it ready to be `.run()`
        """
        pass



from __future__ import annotations

from typing import Optional, TYPE_CHECKING

from ryvencore import Script as ScriptCore

from ironflow.model.flow import Flow

if TYPE_CHECKING:
    from ironflow.model.session import Session


class Script(ScriptCore):
    """
    Wraps `ryvencore.Script.Script` to build our `Flow` class instead.
    """

    def __init__(
        self,
        session: Session,
        title: Optional[str] = None,
        load_data: Optional[dict] = None,
        create_default_logs: bool = True,
    ):
        super().__init__(
            session=session,
            title=title,
            load_data=load_data,
            create_default_logs=create_default_logs,
        )
        del self.flow  # A little inefficient, but I don't want to reproduce init here
        self.flow = Flow(self.session, self)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
For directly controlling node IO (as opposed to giving it to nodes by connecting them to other OI).
"""

from __future__ import annotations

import pickle
import base64
from typing import TYPE_CHECKING, Callable

import ipywidgets as widgets
import numpy as np
from ryvencore.InfoMsgs import InfoMsgs
from traitlets import TraitError

from ironflow.gui.draws_widgets import DrawsWidgets, draws_widgets
from ironflow.model.node import BatchingNode


if TYPE_CHECKING:
    from ipywidgets import DOMWidget
    from ironflow.gui.workflows.screen import WorkflowsGUI
    from ironflow.model.node import Node


def deserialize(data):
    return pickle.loads(base64.b64decode(data))


class NodeController(DrawsWidgets):
    """
    Handles the creation of widgets for manually adjusting node input and viewing node info.
    """

    main_widget_class = widgets.VBox

    def __new__(cls, screen: WorkflowsGUI, *args, **kwargs):
        return super().__new__(cls, *args, **kwargs)

    def __init__(self, screen: WorkflowsGUI, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.screen = screen
        self.node = None
        self._row_height = 30  # px

        self._border = "1px solid black"
        self.widget.layout = widgets.Layout(
            width="50%",
            border="",
            max_height="360px",
            margin="10px",
            padding="5px",
        )

    def draw_for_node(self, node: Node | None) -> None:
        self.clear()
        self.node = node
        if self.node is not None:
            self.draw()

    def update(self):
        if self.node is not None:
            self.draw()

    @draws_widgets
    def draw(self) -> None:
        self.widget.children = [
            self._draw_input_box(),
            self._draw_input_widget(),
            self._draw_info_box(),
        ]
        self.widget.layout.border = self._border
        return self.widget

    def _draw_input_widget(self) -> widgets.Widget:
        try:
            widget = self.node.input_widget(self.screen, self.node).widget
            widget.layout.height = "70px"
            widget.layout.border = "solid 1px blue"
            return widget
        except AttributeError:
            return widgets.Output()

    def _input_field_list(self) -> list[list[widgets.Widget]]:
        input = []
        if hasattr(self.node, "inputs"):
            for i_c, inp in enumerate(self.node.inputs[:]):
                dtype = str(inp.dtype).split(".")[-1]
                disabled = self.node.block_updates or len(inp.connections) > 0
                try:
                    dtype_state = deserialize(inp.data()["dtype state"])
                except TypeError:
                    # `inp.data()` winds up calling `serialize` on `inp.get_val()`
                    # This serialization is a pickle dump, which fails with structures (`Atoms`)
                    # Just gloss over it for now
                    dtype_state = {
                        "val": "Serialization error -- please reconnect an input"
                    }
                if inp.val is None:
                    inp.val = dtype_state["val"]

                try:
                    if dtype_state["batched"]:
                        inp_widget = widgets.Text(
                            f"Batched {dtype}",
                            continuous_update=False,
                            disabled=True,
                        )
                    elif dtype == "Integer":
                        inp_widget = widgets.IntText(
                            value=inp.val,
                            description="",
                            continuous_update=False,
                            disabled=disabled,
                        )
                    elif dtype == "Float":
                        inp_widget = widgets.FloatText(
                            value=inp.val,
                            description="",
                            continuous_update=False,
                            disabled=disabled,
                        )
                    elif dtype == "Boolean":
                        inp_widget = widgets.Checkbox(
                            value=inp.val,
                            indent=False,
                            description="",
                            disabled=disabled,
                        )
                    elif dtype == "Choice":
                        inp_widget = widgets.Dropdown(
                            value=inp.val,
                            options=inp.dtype.items,
                            description="",
                            ensure_option=True,
                            disabled=disabled,
                        )
                    elif dtype == "String" or dtype == "Char":
                        inp_widget = widgets.Text(
                            value=str(inp.val),
                            continuous_update=False,
                            disabled=disabled,
                        )
                    else:
                        inp_widget = widgets.Text(
                            value=str(inp.val), continuous_update=False, disabled=True
                        )
                except TraitError as e:
                    inp_widget = widgets.Label(
                        value="Trait error -- check log and/or change input."
                    )
                    InfoMsgs.write_err(e)

                description = inp.label_str if inp.label_str != "" else inp.type_
                inp_widget.layout.width = "100px"
                inp_widget.observe(self._input_change_i(i_c), names="value")

                batch_button = widgets.ToggleButton(
                    description="Batched",
                    tooltip="Use batches batches of correctly typed data instead of "
                    "instances",
                    layout=widgets.Layout(width="75px"),
                    disabled=inp.dtype is None
                    or not isinstance(inp.node, BatchingNode),
                    value=inp.dtype.batched if hasattr(inp, "dtype") else False,
                )
                batch_button.observe(self._toggle_batching_i(i_c), names="value")

                reset_button = widgets.Button(
                    tooltip="Reset to default",
                    icon="refresh",
                    layout=widgets.Layout(width="30px"),
                    disabled=len(inp.connections) > 0,
                )
                reset_button.on_click(self._input_reset_i(i_c, inp_widget))

                input.append(
                    [widgets.Label(description), inp_widget, batch_button, reset_button]
                )
        return input

    def _input_change_i(self, i_c) -> Callable:
        def input_change(change: dict) -> None:
            # Todo: Test this in exec mode
            self.node.inputs[i_c].update(change["new"])
            # self.node.update(i_c)
            self.screen.redraw_active_flow_canvas()

        return input_change

    def _toggle_batching_i(self, i_c) -> Callable:
        def toggle_batching(change: dict) -> None:
            try:
                InfoMsgs.write(
                    f"Batching for {self.node.title}.{self.node.inputs[i_c].label_str} "
                    f"set to {change['new']}"
                )
                if change["new"]:
                    self.node.inputs[i_c].batch()
                else:
                    self.node.inputs[i_c].unbatch()
                self.screen.redraw_active_flow_canvas()
                self.draw()
            except AttributeError:
                pass

        return toggle_batching

    def _input_reset_i(self, i_c, associated_input_field) -> Callable:
        def input_reset(button: widgets.Button) -> None:
            default = self.node.inputs[i_c].dtype.default
            self.node.inputs[i_c].update(default)
            InfoMsgs.write(
                f"Value for {self.node.title}.{self.node.inputs[i_c].label_str} "
                f"reset to {default}"
            )
            try:
                associated_input_field.value = default
            except TraitError:
                self.screen.update_node_control()
            finally:
                pass
            self.node.update(i_c)
            self.screen.redraw_active_flow_canvas()

        return input_reset

    def _draw_input_box(self) -> widgets.GridBox | widgets.Output:
        input_fields = self._input_field_list()
        n_fields = len(input_fields)
        if n_fields > 0:
            return widgets.GridBox(
                list(np.array(input_fields).flatten()),
                layout=widgets.Layout(
                    grid_template_columns="auto auto auto auto",
                    grid_auto_rows=f"{self._row_height}px",
                    border="solid 1px blue",
                    height=f"{self._box_height(n_fields)}px",
                    # Automatic height like this really should be doable just with the CSS,
                    # but for the life of me I can't get a CSS solution working right -Liam
                ),
            )
        else:
            return widgets.Output()

    def _box_height(self, n_rows: int) -> int:
        return n_rows * self._row_height + 8

    def _draw_info_box(self) -> widgets.VBox:
        glob_id_val = None
        if hasattr(self.node, "GLOBAL_ID"):
            glob_id_val = self.node.GLOBAL_ID
        global_id = widgets.Text(
            value=str(glob_id_val),
            description="GLOBAL_ID:",
            disabled=True,
            layout=widgets.Layout(height=f"{self._row_height}px"),
        )

        title = widgets.Text(
            value=str(self.node.title),
            description="Title:",
            disabled=True,
            layout=widgets.Layout(height=f"{self._row_height}px"),
        )

        info_box = widgets.VBox(
            [title, global_id],
            layout=widgets.Layout(
                height=f"{self._box_height(2)}px",
                border="solid 1px red",
            ),
        )
        return info_box

    def clear(self) -> None:
        self.node = None
        self.widget.children = []
        self.widget.layout.border = ""
        super().clear()



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
The parent canvas for the whole flow representation.
"""

from __future__ import annotations

from time import time
from typing import TYPE_CHECKING, Optional

import ipywidgets as widgets
from ipycanvas import Canvas, hold_canvas
from IPython.display import display

from ironflow.model.port import NodeInput, NodeOutput
from ironflow.gui.workflows.canvas_widgets.base import CanvasWidget
from ironflow.gui.workflows.canvas_widgets.layouts import NodeLayout
from ironflow.gui.workflows.canvas_widgets.nodes import NodeWidget
from ironflow.gui.workflows.canvas_widgets.ports import PortWidget

if TYPE_CHECKING:
    from ironflow.gui.gui import GUI
    from ironflow.gui.workflows.canvas_widgets.base import Number
    from ironflow.gui.workflows.screen import WorkflowsGUI
    from ironflow.model.flow import Flow
    from ironflow.model.node import Node


class FlowCanvas:
    """
    A canvas for representing a particular Ryven script, which is determined at instantiation by the currently active
    gui script.

    Mouse behaviour:
        - Mouse click (down and release) on a node element or any child element selects that element
        - Mouse down, hold, and move on a node element or any child element selects the (parent) node and moves it
        - Mouse click on nothing clears selection
        - Mouse double-click on nothing creates a new node of the type currently selected in the node menu
        - Mouse click on a selected (pressed) port (button) deselects (unpresses) that port (button)
        - Todo: Mouse down, hold, and move on nothing draws a rectangle, everything inside is selected on release

    Keyboard behaviour: Todo
        - ESC: Deselect all.
        - Backspace/Delete:
            - If a node is selected, deletes it
            - If a port is selected, deletes all connections it is part of
    """

    def __init__(self, screen: WorkflowsGUI, flow: Flow):
        self.screen = screen
        self.flow = flow

        self._standard_size = (1800, 800)
        self._zoom_factors = [0.50, 0.75, 1.00, 1.25, 1.50, 1.75, 2.00]
        self._zoom_index = 2
        self._width, self._height = self._get_size()

        self._canvas_color = "black"  # "#584f4e"
        self._connection_style = "white"
        self._connection_width = 3

        self._canvas = Canvas(width=self._width, height=self._height)
        self.canvas_restart()
        self._canvas.layout.width = "100%"
        self._canvas.layout.height = "auto"

        self._canvas.on_mouse_down(self.handle_mouse_down)
        self._canvas.on_mouse_up(self.handle_mouse_up)
        self._canvas.on_mouse_move(self.handle_mouse_move)
        self._canvas.on_key_down(self.handle_keyboard_event)

        self.output = widgets.Output(layout={"border": "1px solid black"})

        self.objects_to_draw = []
        self.connections = []

        self.x = 0
        self.y = 0
        self._x_move_anchor = None
        self._y_move_anchor = None

        self._last_selected_object = None

        self._mouse_is_down = False
        self._last_mouse_down = time()
        self._double_click_speed = (
            0.25  # In seconds. Todo: Put this in a config somewhere
        )

        self._object_to_gui_dict = {}
        self._highlighted_ports: list[PortWidget] = []

    @property
    def canvas(self):
        return self._canvas

    @property
    def flow_canvas(self) -> FlowCanvas:
        return self

    @property
    def gui(self) -> GUI:
        return self.screen.gui

    @property
    def title(self) -> str:
        return self.flow.script.title

    def display(self):
        self.output.clear_output()
        with self.output:
            display(self.canvas)

    def draw_connection(self, port_1: int, port_2: int) -> None:
        out = self._object_to_gui_dict[port_1]
        inp = self._object_to_gui_dict[port_2]

        canvas = self._canvas
        canvas.stroke_style = self._connection_style
        canvas.line_width = self._connection_width
        canvas.move_to(out.x, out.y)
        canvas.line_to(inp.x, inp.y)
        canvas.stroke()

    def _built_object_to_gui_dict(self) -> None:
        self._object_to_gui_dict = {}
        for n in self.objects_to_draw:
            self._object_to_gui_dict[n.node] = n
            for p in n.objects_to_draw:
                if isinstance(p, PortWidget):
                    self._object_to_gui_dict[p.port] = p

    def canvas_restart(self) -> None:
        self._canvas.clear()
        self._canvas.fill_style = self._canvas_color
        self._canvas.fill_rect(0, 0, self._width, self._height)

    def handle_keyboard_event(self, key: str, shift_key, ctrl_key, meta_key) -> None:
        pass  # Todo

    def deselect_all(self) -> None:
        [o.deselect() for o in self.objects_to_draw]
        self._last_selected_object = None
        self.redraw()

    def handle_mouse_down(self, x: Number, y: Number):
        self._mouse_is_down = True
        self._x_move_anchor = x
        self._y_move_anchor = y

        now = time()
        time_since_last_click = now - self._last_mouse_down
        self._last_mouse_down = now

        sel_object = self.get_element_at_xy(x, y)
        last_object = self._last_selected_object

        if sel_object is None:
            if last_object is not None:
                last_object.deselect()
            elif time_since_last_click < self._double_click_speed:
                self.add_node(x, y, self.screen.new_node_class)
                self._built_object_to_gui_dict()
        else:
            if (
                sel_object == last_object
                and time_since_last_click < self._double_click_speed
            ):
                sel_object = sel_object.on_double_click()
            else:
                sel_object = sel_object.on_click(last_object)

        self._last_selected_object = sel_object

        self.redraw()

    def handle_mouse_up(self, x: Number, y: Number):
        self._mouse_is_down = False

    def get_element_at_xy(self, x_in: Number, y_in: Number) -> CanvasWidget | None:
        for o in self.objects_to_draw:
            if o.is_here(x_in, y_in):
                return o.get_element_at_xy(x_in, y_in)
        return None

    def get_selected_objects(self) -> list[CanvasWidget]:
        return [o for o in self.objects_to_draw if o.selected]

    def handle_mouse_move(self, x: Number, y: Number) -> None:
        if self._mouse_is_down:
            selected_objects = self.get_selected_objects()
            if len(selected_objects) > 0:
                with hold_canvas(self._canvas):
                    [o.set_x_y(x - self.x, y - self.y) for o in selected_objects]
                    self.redraw()
            else:
                self.x += x - self._x_move_anchor
                self.y += y - self._y_move_anchor
                self._x_move_anchor = x
                self._y_move_anchor = y
                self.redraw()

    def redraw(self) -> None:
        with hold_canvas(self._canvas):
            self.canvas_restart()
            [o.draw() for o in self.objects_to_draw]
            for c in self.flow.connections:
                self.draw_connection(c.inp, c.out)
        self.screen.update_node_presenter()

    def load_node(self, x: Number, y: Number, node: Node) -> NodeWidget:
        layout = NodeLayout()

        if hasattr(node, "main_widget_class") and node.main_widget_class is not None:
            if isinstance(node.main_widget_class, str):
                node_class = eval(node.main_widget_class)
            elif issubclass(node.main_widget_class, NodeWidget):
                node_class = node.main_widget_class
            else:
                raise TypeError(
                    f"main_widget_class {node.main_widget_class} not recognized"
                )
        else:
            node_class = NodeWidget
        s = node_class(x=x, y=y, parent=self, layout=layout, node=node)

        self.objects_to_draw.append(s)
        return s

    def add_node(self, x: Number, y: Number, node: Node):
        n = self.flow.create_node(node)
        self.load_node(x, y, n)

        self.redraw()

    def delete_selected(self) -> None:
        for o in self.objects_to_draw:
            if o.selected:
                o.delete()
        self.redraw()

    def _get_size(self) -> tuple[int, int]:
        scale = self._zoom_factors[self._zoom_index]
        return int(self._standard_size[0] * scale), int(self._standard_size[1] * scale)

    def _zoom(self, zoom_index) -> None:
        if zoom_index != self._zoom_index:
            self._zoom_index = zoom_index
            self._width, self._height = self._get_size()
            self._canvas.width = self._width
            self._canvas.height = self._height
            self.redraw()
        else:
            self.screen.print("Zoom limit reached")

    def zoom_in(self) -> None:
        self._zoom(max(self._zoom_index - 1, 0))

    def zoom_out(self) -> None:
        self._zoom(min(self._zoom_index + 1, len(self._zoom_factors) - 1))

    def highlight_compatible_ports(self, selected: PortWidget):
        if selected.port.otype is None:
            return

        compatible_port_widgets = self._get_port_widgets_ontologically_compatible_with(
            selected.port
        )

        for port_widget in compatible_port_widgets:
            port_widget.highlight()
        self._highlighted_ports = compatible_port_widgets

    def _get_port_widgets_ontologically_compatible_with(self, port):
        if isinstance(port, NodeInput):
            input_tree = port.otype.get_source_tree(
                additional_requirements=port.get_downstream_requirements()
            )
            return [
                subwidget
                for subwidget in self._port_widgets
                if isinstance(subwidget.port, NodeOutput)
                and subwidget.port.all_connections_found_in(input_tree)
            ]
        elif isinstance(port, NodeOutput):
            return [
                subwidget
                for subwidget in self._port_widgets
                if subwidget.port.otype is not None  # Progressively expensive checks
                and port.otype in subwidget.port.otype.get_sources()
                and subwidget.port.workflow_tree_contains_connections_of(port)
            ]
        else:
            raise TypeError(
                f"Expected a {NodeInput} or {NodeOutput} but got {type(port)}"
            )

    @property
    def _port_widgets(self):
        return [
            subwidget
            for node_widget in self.objects_to_draw
            for subwidget in node_widget.objects_to_draw
            if isinstance(subwidget, PortWidget)
        ]

    def clear_port_highlighting(self):
        for port_widget in self._highlighted_ports:
            port_widget.dehighlight()



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
ipywidgets holds each instantiated widget independently, and they aren't released to the
garbage collector until their `close` method has been called and there are no further
references to their instance.

If we're not careful, this can lead to a massive bloat of the registered widgets and
memory strain. Here, we introduce a abstract class to help manage the closure of the
widgets we create.
"""

from __future__ import annotations
from abc import ABC, ABCMeta, abstractmethod
from typing import Callable

import ipywidgets as widgets


class PostCaller(type):
    """
    Inspired by [@Cam.Davidson.Pilon](https://stackoverflow.com/questions/795190/how-to-perform-common-post-initialization-tasks-in-inherited-classes)
    """

    def __call__(cls, *args, **kwargs):
        obj = type.__call__(cls, *args, **kwargs)
        obj.__post__()
        return obj


class AbstractPostCaller(PostCaller, ABCMeta):
    pass


def draws_widgets(fnc: Callable) -> Callable:
    """
    A decorator for any class methods that instantiate new widgets outside init.
    """

    def wrapper(self, *args, **kwargs):
        n_widgets_i = len(widgets.Widget.widgets)
        result = fnc(self, *args, **kwargs)
        n_widgets_drawn = len(widgets.Widget.widgets) - n_widgets_i
        if n_widgets_drawn > 0:
            self._drawn_widgets += list(widgets.Widget.widgets.values())[
                -n_widgets_drawn:
            ]
        return result

    return wrapper


class DrawsWidgets(ABC, metaclass=AbstractPostCaller):
    """
    A mixin for classes that instantiate at least one ipywidgets widget during their
    `__init__` and _only_ instantiate other widgets in methods decorated with the
    `draws_widgets` decorator.

    All instantiated widgets are logged and can then be destroyed with the `clear`
    or `close` methods for those widgets instantiated since the last clear call or all
    instantiated widgets, respectively.

    Children will instantiate one `widget` attribute, which will be of type
    `main_widget_class`, which must be of type `ipywidgets.Box` (i.e. it takes a list of
    `children` as its first argument.)

    Note: Child classes will need to additionally define `__new__` to pull out all args
        and kwargs that are explicitly passed into the init, and pass on only
        `super().__new__(cls, *args, **kwargs)` upstream.
    """

    main_widget_class: type[widgets.Box] = widgets.Box

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls, *args, **kwargs)
        obj.__n_widgets_at_init = len(widgets.Widget.widgets)
        return obj

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.widget = self.main_widget_class([])
        self._init_widgets: list[widgets.Widget] = None
        self._drawn_widgets: list[widgets.Widget] = []

    def __post__(self):
        n_init_widgets = len(widgets.Widget.widgets) - self.__n_widgets_at_init
        self._init_widgets = list(widgets.Widget.widgets.values())[-n_init_widgets:]

    def clear(self):
        """
        Delete all widgets instantiated between the end of init and the last clear call.
        """
        self._delete_widget_list(self._drawn_widgets)

    @staticmethod
    def _delete_widget_list(widget_list: list[widgets.Widget]):
        while widget_list:
            w = widget_list.pop()
            w.close()
            del w

    def __del__(self):
        self.close()
        try:
            super().__del__()
        except AttributeError:
            pass

    def close(self):
        """
        Call clear, then delete all widgets instantiated in init.
        """
        self.clear()
        self._delete_widget_list(self._init_widgets)



"""
Stuff that doesn't fit anywhere yet.
"""

from __future__ import annotations

from IPython.core.display import HTML


def display_string(string):
    """
    Format a python string to look nice in an IPython.display.display context.

    If we just pass a string, `display` doesn't resolve newlines.
    If we pass a `print`ed string, `display` also shows the `None` value returned by `print`
    So we use this ugly hack.
    """
    return HTML(
        string.replace("\n", "<br>").replace("\t", "&emsp;").replace(" ", "&nbsp;")
    )



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Button-like canvas widgets.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional

from ironflow.model.port import NodeInput, NodeOutput
from ironflow.gui.workflows.canvas_widgets.base import CanvasWidget, HideableWidget
from ironflow.gui.workflows.canvas_widgets.layouts import ButtonLayout

if TYPE_CHECKING:
    from ironflow.gui.workflows.canvas_widgets.base import Number
    from ironflow.gui.workflows.canvas_widgets.nodes import NodeWidget
    from ironflow.model import NodePort


class ButtonWidget(CanvasWidget, ABC):
    def __init__(
        self,
        x: Number,
        y: Number,
        parent: NodeWidget,
        layout: ButtonLayout,
        selected: bool = False,
        title: str = "Button",
        pressed: Optional[bool] = False,
    ):
        super().__init__(x, y, parent, layout, selected)
        self.title = title
        self.pressed = pressed

    def on_click(
        self, last_selected_object: Optional[CanvasWidget]
    ) -> CanvasWidget | None:
        if self.pressed:
            self.unpress()
        else:
            self.press()
        self.deselect()
        return last_selected_object

    def press(self):
        self.pressed = True
        self.on_pressed()
        # Ok, now here is a weird hack:
        self.screen.update_node_control()
        # This way, if a button causes some change in node-state, it gets reflected in the node_controller,
        # e.g. if we `run` or `remove` some job, but a nodestream node is taking that job as input.
        # Performance and usability hits are minimal, but from a maintenance perspective it is super bad, confusing,
        # and hard to maintain, so it would be much nicer in the future to find a way to hook the node_controller right
        #  onto input updates on the nodes

    def unpress(self):
        self.pressed = False
        self.on_unpressed()

    @abstractmethod
    def on_pressed(self):
        pass

    @abstractmethod
    def on_unpressed(self):
        pass

    def draw_shape(self) -> None:
        self.canvas.fill_style = (
            self.layout.pressed_color if self.pressed else self.layout.background_color
        )
        self.canvas.fill_rect(
            self.x,
            self.y,
            self.width,
            self.height,
        )

    def draw_title(self) -> None:
        self.canvas.font = self.layout.font_string
        self.canvas.fill_style = self.layout.font_color
        x = self.x + (self.width * 0.1)
        y = self.y + (self.height * 0.05) + self.layout.font_size
        self.canvas.fill_text(self.title, x, y)


class RepresentButtonWidget(ButtonWidget):
    def __init__(
        self,
        x: Number,
        y: Number,
        parent: NodeWidget,
        layout: ButtonLayout,
        selected: bool = False,
        title="SHOW",
    ):
        super().__init__(x, y, parent, layout, selected, title=title)

    def on_pressed(self):
        self.screen.open_node_presenter(self.parent)

    def on_unpressed(self):
        self.screen.close_node_presenter()


class ExpandCollapseButtonWidget(ButtonWidget, HideableWidget, ABC):
    def __init__(
        self,
        x: Number,
        y: Number,
        parent: NodeWidget,
        layout: ButtonLayout,
        selected: bool = False,
        pressed: bool = False,
        visible: bool = True,
        title: Optional[str] = None,
        size: Optional[Number] = None,
    ):
        if size is not None:
            layout.width = size
            layout.height = size
        layout.background_color = parent.node.color
        layout.pressed_color = parent.node.color

        ButtonWidget.__init__(
            self,
            x=x,
            y=y,
            parent=parent,
            layout=layout,
            selected=selected,
            title=title,
            pressed=pressed,
        )
        HideableWidget.__init__(
            self,
            x=x,
            y=y,
            parent=parent,
            layout=layout,
            selected=selected,
            title=title,
            visible=visible,
        )

    def on_pressed(self):
        self.hide()

    def on_unpressed(self):
        self.show()

    def draw_shape(self) -> None:
        self.canvas.fill_style = (
            self.layout.pressed_color if self.pressed else self.layout.background_color
        )
        self.canvas.fill_polygon(self._points)

    @property
    @abstractmethod
    def _points(self) -> list[tuple[Number, Number]]:
        pass


class ExpandButtonWidget(ExpandCollapseButtonWidget):
    @property
    def _points(self) -> list[tuple[Number, Number]]:
        return [
            (self.x, self.y),
            (self.x + self.width, self.y),
            (self.x + 0.5 * self.width, self.y + self.height),
        ]

    def on_pressed(self):
        super().on_pressed()
        self.parent.expand_io()


class CollapseButtonWidget(ExpandCollapseButtonWidget):
    @property
    def _points(self) -> list[tuple[Number, Number]]:
        return [
            (self.x, self.y + self.height),
            (self.x + 0.5 * self.width, self.y),
            (self.x + self.width, self.y + self.height),
        ]

    def on_pressed(self):
        super().on_pressed()
        self.parent.collapse_io()


class ExecButtonWidget(ButtonWidget):
    def __init__(
        self,
        x: Number,
        y: Number,
        parent: NodeWidget,
        layout: ButtonLayout,
        port: NodePort,
        selected: bool = False,
        title: str = "Exec",
        pressed: Optional[bool] = False,
    ):
        super().__init__(
            x=x,
            y=y,
            parent=parent,
            layout=layout,
            selected=selected,
            title=port.label_str if port.label_str != "" else title,
            pressed=pressed,
        )
        self.port = port

    def on_pressed(self):
        self.flow_canvas.redraw()  # Draw the button pressed
        if isinstance(self.port, NodeInput):
            self.port.update()
        elif isinstance(self.port, NodeOutput):
            self.port.exec()
        self.unpress()  # At the next re-draw, the button will appear unpressed again,
        # i.e. at the end of the mouse-down that triggered this

    def on_unpressed(self):
        pass



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Initially wrapped dtypes to avoid mutable defaults, but after that expanded to
facilitate strict type checking when flow connections are made, and to add batching.

Node ports were overridden so that by default they come with an `Untyped` dtype, and
always have a batching flag.

Further, the dtypes are broken down into broad categories of `Data`, `List`, and
`Choice` with different behaviours under regular and batched conditions.

Warning:
    Any additional types defined here later need to be added to the list in
    `DType.from_str` to work with (de)serialization.

Implementation of Dtypes changes in ryvencore v0.4, so this file may be short-lived.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Optional

import numpy as np
from ryvencore.dtypes import DType as DTypeCore


def isiterable(obj):
    try:
        iter(obj)
        return True
    except TypeError:
        return False


def other_classes_are_subset(other, reference):
    return all(any(issubclass(o, ref) for ref in reference) for o in other)


class DType(DTypeCore, ABC):
    def __init__(
        self,
        default,
        bounds: tuple = None,
        doc: str = "",
        _load_state=None,
        valid_classes=None,
        allow_none=False,
        batched=False,
    ):
        super().__init__(
            default=default,
            bounds=bounds,
            doc=doc,
            _load_state=_load_state,
        )
        if _load_state is None:
            if isinstance(valid_classes, list):
                self.valid_classes = list(valid_classes)
            elif valid_classes is not None:
                self.valid_classes = [valid_classes]
            else:
                self.valid_classes = []
            self.allow_none = allow_none
            self.batched = batched
        self.add_data("valid_classes")
        self.add_data("allow_none")
        self.add_data("batched")

    @staticmethod
    def from_str(s):
        # Load local dtypes, not ryven dtypes
        for DTypeClass in [
            Boolean,
            Choice,
            Data,
            Float,
            Integer,
            List,
            String,
            Untyped,
        ]:
            if s == "DType." + DTypeClass.__name__:
                return DTypeClass

        return None

    def _classes_are_subset(self, other_classes):
        return other_classes_are_subset(other_classes, self.valid_classes)

    def accepts(self, other: DType | Any | None):
        if isinstance(other, DType):
            return self._accepts_dtype(other)
        else:
            return self._accepts_instance(other)

    @abstractmethod
    def _accepts_dtype(self, other: DType):
        pass

    def _accepts_instance(self, val: Any):
        if self.batched:
            return self._batch_accepts_instance(val)
        else:
            return self._accepts_none(val) or self._accepts_non_none_instance(val)

    @abstractmethod
    def _batch_accepts_instance(self, val: Any):
        pass

    def _accepts_none(self, val: Any):
        return val is None and self.allow_none

    @abstractmethod
    def _accepts_non_none_instance(self, val: Any):
        pass

    def valid_val(self, val: Any):
        return self._accepts_instance(val)

    def _surprise_none_possible(self, other: DType):
        return other.allow_none and not self.allow_none


class Untyped(DType):
    """
    Untyped data always performs an instance-check when used as input and when it's
    used as output, other input nodes always perform an instance-check against it.
    That means it can't be used to pre-wire a graph that has missing data.

    When Untyped as an input receives output connections...
    Normally:
        - Accept anything.
    When batched:
        - Accept any value that is iterable.

    Untyped is always valid unless the value is None and None is not allowed.
    """

    def __init__(
        self,
        doc: str = "",
        _load_state=None,
        batched=False,
    ):
        super().__init__(
            default=None,
            bounds=None,
            doc=doc,
            _load_state=_load_state,
            valid_classes=None,
            allow_none=True,
            batched=batched,
        )

    def _accepts_dtype(self, other: DType):
        raise ValueError(
            f"Match checks to {self.__class__.__name__} should always be done by "
            f"value, not by dtype. Please contact a package maintainer and explain how"
            f"you got to this error."
        )

    def _accepts_non_none_instance(self, val: Any):
        return True

    def _batch_accepts_instance(self, val: Any):
        return hasattr(val, "__iter__")


class Data(DType):
    """
    For most types of data.

    When Data as an input receives output connections...
    Normally:
        - Data output: output valid classes must be a subset of input valid classes, and
            one of input or output classes must inherit from the other (or be the same).
        - Untyped output: output value must be an instance of valid classes
        - All else: Fail.
    When batched:
        - Batched Data output: same as the unbatched case, but now both are batched.
        - Untyped output: output value must be iterable and each element must be an
            instance of valid classes.
        - List output: output valid classes must be a subset of input valid classes.
        - All else: Fail.

    Data is valid when the value is an instance of the valid classes, or is None and
    None is allowed.
    """

    def __init__(
        self,
        default=None,
        size: str = "m",
        doc: str = "",
        _load_state=None,
        valid_classes=None,
        allow_none=False,
        batched=False,
    ):
        """
        size: 's' / 'm' / 'l'
        """
        self.size = size
        super().__init__(
            default=default,
            doc=doc,
            _load_state=_load_state,
            valid_classes=valid_classes,
            allow_none=allow_none,
            batched=batched,
        )
        self.add_data("size")

    def _accepts_dtype(self, other: DType):
        if isinstance(other, Untyped):
            raise ValueError(
                f"Match checks against {Untyped.__class__.__name__} should always be "
                f"done by value, not by dtype. Please contact a package maintainer and "
                f"explain how you got to this error."
            )
        elif (
            isinstance(other, self.__class__) or isinstance(self, other.__class__)
        ) and other.batched == self.batched:
            return self._classes_are_subset(
                other.valid_classes
            ) and not self._surprise_none_possible(other)
        elif self.batched and isinstance(other, List) and not other.batched:
            return self._classes_are_subset(other.valid_classes)
        else:
            return False

    def _accepts_non_none_instance(self, val: Any):
        return any(isinstance(val, c) for c in self.valid_classes)

    def _batch_accepts_instance(self, val: Any):
        if hasattr(val, "__iter__"):
            if any(v is None for v in val) and not self.allow_none:
                return False
            else:
                return self._classes_are_subset(
                    set([type(v) for v in val if v is not None])
                )
        else:
            return False


class Integer(Data):
    def __init__(
        self,
        default: int = 0,
        bounds: tuple = None,
        doc: str = "",
        _load_state=None,
        valid_classes=None,
        allow_none=False,
        batched=False,
    ):
        self.bounds = bounds
        super().__init__(
            default=default,
            doc=doc,
            _load_state=_load_state,
            valid_classes=[int, np.integer] if valid_classes is None else valid_classes,
            allow_none=allow_none,
            batched=batched,
        )
        self.add_data("bounds")


class Float(Data):
    def __init__(
        self,
        default: float = 0.0,
        bounds: tuple = None,
        decimals: int = 10,
        doc: str = "",
        _load_state=None,
        valid_classes=None,
        allow_none=False,
        batched=False,
    ):
        self.bounds = bounds
        self.decimals = decimals
        super().__init__(
            default=default,
            doc=doc,
            _load_state=_load_state,
            valid_classes=(
                [float, np.floating] if valid_classes is None else valid_classes
            ),
            allow_none=allow_none,
            batched=batched,
        )
        self.add_data("bounds")
        self.add_data("decimals")


class Boolean(Data):
    def __init__(
        self,
        default: bool = False,
        doc: str = "",
        _load_state=None,
        valid_classes=None,
        allow_none=False,
        batched=False,
    ):
        super().__init__(
            default=default,
            doc=doc,
            _load_state=_load_state,
            valid_classes=[bool, np.bool_] if valid_classes is None else valid_classes,
            allow_none=allow_none,
            batched=batched,
        )


class String(Data):
    def __init__(
        self,
        default: str = "",
        doc: str = "",
        _load_state=None,
        valid_classes=None,
        allow_none=False,
        batched=False,
    ):
        super().__init__(
            default=default,
            doc=doc,
            _load_state=_load_state,
            valid_classes=[str, np.str_] if valid_classes is None else valid_classes,
            allow_none=allow_none,
            batched=batched,
        )


class Choice(DType):
    """
    Data that must be chosen from among a list of items.

    When Choice as an input receives output connections...
    Normally:
        - Data output: output valid classes must be a subset.
        - Untyped output: output value must be in the items list.
        - All else: Fail.
    When batched:
        - Batched Data output: output valid classes must be a subset.
        - List output: output valid classes must be a subset.
        - Untyped: output value must be iterable, and each element must be in the items
            list.
        - All else: Fail.

    Choice is valid when the value is in the items list, or value is None and None is
    allowed.

    Note that when making Data (or List) connections, the connection may be allowed
    but still result in an invalid value state (in cases where the output value does
    not match the input items list).
    """

    def __init__(
        self,
        default=None,
        items: Optional[list] = None,
        doc: str = "",
        _load_state=None,
        valid_classes=None,
        allow_none=False,
        batched=False,
    ):
        self.items = items if items is not None else []
        super().__init__(
            default=default,
            doc=doc,
            _load_state=_load_state,
            valid_classes=valid_classes,
            allow_none=allow_none,
            batched=batched,
        )
        self.add_data("items")

    def _accepts_dtype(self, other: DType):
        # TODO: Temporary code duplication while splitting Data and Choice
        if isinstance(other, Untyped):
            raise ValueError(
                f"Match checks against {Untyped.__class__.__name__} should always be "
                f"done by value, not by dtype. Please contact a package maintainer and "
                f"explain how you got to this error."
            )
        if self.batched:
            dtype_ok = (isinstance(other, List) and not other.batched) or (
                isinstance(other, Data) and other.batched
            )
            classes_ok = self._classes_are_subset(other.valid_classes)
            return dtype_ok and classes_ok and not self._surprise_none_possible(other)
        elif isinstance(other, Data) and not other.batched:
            return self._classes_are_subset(
                other.valid_classes
            ) and not self._surprise_none_possible(other)
        else:
            return False

    def _accepts_non_none_instance(self, val: Any):
        return val in self.items

    def _batch_accepts_instance(self, val: Any):
        return isinstance(val, (list, np.ndarray)) and all(
            self._accepts_non_none_instance(v) for v in val
        )


class List(DType):
    """
    Data that is explicitly iterable.

    When List as an input receives output connections...
    Normally:
        - List output: output valid classes must be a subset.
        - Batched Data output: output valid classes must be a subset.
        - Untyped output: output value must be iterable and each element must be an
            instance of valid classes.
        - All else: Fail.
    When batched:
        - Batched List output: output valid classes must be a subset.
        - Untyped: output value must be iterable, each element must be iterable, and
            each element's element must be an instance of a valid class.
        - All else: Fail.

    List is valid when the value is iterable and all elements are instances of the
    valid classes, or the value is None and None is allowed.

    Note: `allow_none` in this case determines whether the _entire dtype value_ may be
        `None`. If you want to specify that the list-like object itself may _contain_
        `None` values, add `type(None)` to the `valid_classes`.
    """

    def __init__(
        self,
        default: Optional[list] = None,
        doc: str = "",
        _load_state=None,
        valid_classes=None,
        allow_none=False,
        batched=False,
    ):
        super().__init__(
            default=default,
            doc=doc,
            _load_state=_load_state,
            valid_classes=list if valid_classes is None else valid_classes,
            allow_none=allow_none,
            batched=batched,
        )

    def _accepts_dtype(self, other: DType):
        if self.batched:
            return (
                isinstance(other, List)
                and other.batched
                and self._classes_are_subset(other.valid_classes)
            )
        elif isinstance(other, List) or (isinstance(other, Data) and other.batched):
            # TODO: Only other unbatched lists should be accepted to conform to spec
            #       At the moment, this is a very useful bug, since it lets us pass
            #       batched data to the `Transpose` and `Slice` nodes to modify them
            #       The correct fix is to introduce a new Matrix DType, of which
            #       List is a special case
            return self._classes_are_subset(other.valid_classes)
        else:
            return False

    def _accepts_non_none_instance(self, val: Any):
        return isiterable(val) and all(
            any(isinstance(v, c) for c in self.valid_classes) for v in val
        )

    def _batch_accepts_instance(self, val: Any):
        return isiterable(val) and all(
            self._accepts_none(v) or self._accepts_non_none_instance(v) for v in val
        )



from ironflow.node_tools import (
    dtypes,
    Node,
    NodeInputBP,
    NodeOutputBP,
    PlaceholderWidgetsContainer,
)

widgets = PlaceholderWidgetsContainer()


class NodeBase(Node):
    pass


class GetVar_Node(NodeBase):
    """Gets the value of a script variable"""

    version = "v0.1"

    title = "get var"
    init_inputs = [
        NodeInputBP(dtype=dtypes.String()),
    ]
    init_outputs = [NodeOutputBP(label="val")]
    color = "#c69a15"

    def __init__(self, params):
        super().__init__(params)

        self.var_name = ""
        self.temp_var_val = None

    def place_event(self):
        self.update()

    def view_place_event(self):
        self.var_name = self.input(0)

    def update_event(self, input_called=-1):
        if self.input(0) != self.var_name:
            if self.var_name != "":  # disconnect old var val update connection
                self.unregister_var_receiver(self.var_name, self.var_val_changed)

            self.var_name = self.input(0)

            # create new var update connection
            self.register_var_receiver(self.var_name, self.var_val_changed)

        self.set_output_val(0, self.get_var_val(self.var_name))

    def var_val_changed(self, name, val):
        self.set_output_val(0, val)


class Result_Node(NodeBase):
    """Simply shows a value converted to str"""

    version = "v0.1"

    title = "result"
    init_inputs = [
        NodeInputBP(type_="data"),
    ]
    main_widget_class = widgets.Result_Node_MainWidget
    main_widget_pos = "between ports"
    color = "#c69a15"

    def __init__(self, params):
        super().__init__(params)
        self.val = None

    def place_event(self):
        self.update()

    def view_place_event(self):
        self.main_widget().show_val(self.val)

    def update_event(self, input_called=-1):
        self.val = self.input(0)
        if self.session.gui:
            self.main_widget().show_val(self.val)


class Val_Node(NodeBase):
    """Evaluates a string from the input field"""

    version = "v0.1"

    title = "val"
    init_inputs = [
        NodeInputBP(dtype=dtypes.Data(size="s")),
    ]
    init_outputs = [
        NodeOutputBP(type_="data"),
    ]
    style = "small"
    color = "#c69a15"

    def __init__(self, params):
        super().__init__(params)

        self.display_title = ""
        self.actions["edit val via dialog"] = {"method": self.action_edit_via_dialog}
        self.val = None

    def place_event(self):
        self.update()

    def update_event(self, input_called=-1):
        self.val = self.input(0)
        self.set_output_val(0, self.val)

    def action_edit_via_dialog(self):
        return

        # from ..EditVal_Dialog import EditVal_Dialog
        #
        # val_dialog = EditVal_Dialog(parent=None, init_val=self.val)
        # accepted = val_dialog.exec_()
        # if accepted:
        #     self.main_widget().setText(str(val_dialog.get_val()))
        #     self.update()

    def get_current_var_name(self):
        return self.input(0)

    def get_state(self):
        return {"val": self.val}  # self.main_widget().get_val()

    def set_state(self, data, version):
        self.val = data["val"]

        if version is None:
            self.display_title = ""

            self.create_input_dt(dtype=dtypes.Data(size="s"))

            # the old version didn't use a dtype
            self.inputs[0].dtype.val = self.val
            self.inputs[0].update(self.val)


class SetVar_Node(NodeBase):
    """Sets the value of a script variable"""

    version = "v0.1"

    title = "set var"
    init_inputs = [
        NodeInputBP(type_="exec"),
        NodeInputBP(dtype=dtypes.String(), label="var"),
        NodeInputBP(dtype=dtypes.Data(size="s"), label="val"),
    ]
    init_outputs = [
        NodeOutputBP(type_="exec"),
        NodeOutputBP(type_="data", label="val"),
    ]
    style = "normal"
    color = "#c69a15"

    def __init__(self, params):
        super().__init__(params)

        self.actions["make passive"] = {"method": self.action_make_passive}
        self.active = True

        self.var_name = ""
        self.num_vars = 1

    def update_event(self, input_called=-1):
        if self.active and input_called == 0:
            if self.set_var_val(self.input(1), self.input(2)):
                self.set_output_val(1, self.input(2))
            self.exec_output(0)

        elif not self.active:
            self.var_name = self.input(0)
            if self.set_var_val(self.input(0), self.input(1)):
                self.set_output_val(0, self.get_var_val(self.var_name))

    def action_make_passive(self):
        self.active = False
        self.delete_input(0)
        self.delete_output(0)
        del self.actions["make passive"]
        self.actions["make active"] = {"method": self.action_make_active}

    def action_make_active(self):
        self.active = True
        self.create_input(type_="exec", pos=0)
        self.create_output(type_="exec", pos=0)
        del self.actions["make active"]
        self.actions["make passive"] = {"method": self.action_make_passive}

    def get_state(self):
        return {"active": self.active}

    def set_state(self, data, version):
        self.active = data["active"]


class SetVarsPassive_Node(NodeBase):
    """Sets the values of multiple script variables"""

    version = "v0.1"

    title = "set vars passive"
    init_inputs = []
    init_outputs = []
    style = "normal"
    color = "#c69a15"

    def __init__(self, params):
        super().__init__(params)

        self.actions["add var input"] = {"method": self.add_var_input}

        self.num_vars = 0

    def place_event(self):
        if self.num_vars == 0:
            self.add_var_input()

    def add_var_input(self):
        self.create_input_dt(label="var", dtype=dtypes.String(size="l"))
        self.create_input_dt(label="val", dtype=dtypes.Data(size="l"))
        self.num_vars += 1

        self.actions[f"remove var {self.num_vars}"] = {
            "method": self.remove_var_input,
            "data": self.num_vars,
        }

    def remove_var_input(self, number):
        self.delete_input((number - 1) * 2)
        self.delete_input((number - 1) * 2)
        self.num_vars -= 1
        self.rebuild_remove_actions()

    def rebuild_remove_actions(self):
        remove_keys = []
        for k, v in self.actions.items():
            if k.startswith("remove var"):
                remove_keys.append(k)

        for k in remove_keys:
            del self.actions[k]

        for i in range(self.num_vars):
            self.actions[f"remove var {i+1}"] = {
                "method": self.remove_var_input,
                "data": i + 1,
            }

    def update_event(self, input_called=-1):
        var_names = [self.input(i) for i in range(0, len(self.inputs), 2)]
        values = [self.input(i) for i in range(1, len(self.inputs), 2)]

        for i in range(len(var_names)):
            self.set_var_val(var_names[i], values[i])

    def get_state(self):
        return {"num vars": self.num_vars}

    def set_state(self, data, version):
        self.num_vars = data["num vars"]



from ironflow.node_tools import dtypes, Node, NodeInputBP, NodeOutputBP


class OperatorNodeBase(Node):
    version = "v0.0"

    init_inputs = [
        NodeInputBP(dtype=dtypes.Data(size="s")),
        NodeInputBP(dtype=dtypes.Data(size="s")),
    ]

    init_outputs = [
        NodeOutputBP(),
    ]

    style = "small"

    def __init__(self, params):
        super().__init__(params)

        self.num_inputs = 0
        self.actions["add input"] = {"method": self.add_operand_input}

    def place_event(self):
        for i in range(len(self.inputs)):
            self.register_new_operand_input(i)

    def add_operand_input(self):
        self.create_input_dt(dtype=dtypes.Data(size="s"))
        self.register_new_operand_input(self.num_inputs)
        self.update()

    def remove_operand_input(self, index):
        self.delete_input(index)
        self.num_inputs -= 1
        # del self.actions[f'remove input {index}']
        self.rebuild_remove_actions()
        self.update()

    def register_new_operand_input(self, index):
        self.actions[f"remove input {index}"] = {
            "method": self.remove_operand_input,
            "data": index,
        }
        self.num_inputs += 1

    def rebuild_remove_actions(self):
        remove_keys = []
        for k, v in self.actions.items():
            if k.startswith("remove input"):
                remove_keys.append(k)

        for k in remove_keys:
            del self.actions[k]

        for i in range(self.num_inputs):
            self.actions[f"remove input {i}"] = {
                "method": self.remove_operand_input,
                "data": i,
            }

    def update_event(self, inp=-1):
        self.set_output_val(
            0, self.apply_op([self.input(i) for i in range(len(self.inputs))])
        )

    def apply_op(self, elements: list):
        return None


# LOGIC -------------------------------------


class LogicNodeBase(OperatorNodeBase):
    color = "#f58142"


class NOT_Node(LogicNodeBase):
    title = "not"

    def apply_op(self, elements: list):
        return all([not bool(e) for e in elements])


class AND_Node(LogicNodeBase):
    title = "and"

    def apply_op(self, elements: list):
        return all(elements)


class NAND_Node(LogicNodeBase):
    title = "nand"

    def apply_op(self, elements: list):
        return not all(elements)


class OR_Node(LogicNodeBase):
    title = "or"

    def apply_op(self, elements: list):
        return any(elements)


class NOR_Node(LogicNodeBase):
    title = "nor"

    def apply_op(self, elements: list):
        return not any(elements)


class XOR_Node(LogicNodeBase):
    title = "xor"

    def apply_op(self, elements: list):
        # XOR definition for unbound number of operands:
        #   odd number of operands must be true
        return len(list(filter((lambda x: bool(x)), elements))) % 2 != 0


class XNOR_Node(LogicNodeBase):
    title = "xnor"

    def apply_op(self, elements: list):
        # XNOR definition for unbound number of operands:
        #   even number of operands must be true
        return len(list(filter((lambda x: bool(x)), elements))) % 2 == 0


logic_nodes = [
    NOT_Node,
    AND_Node,
    NAND_Node,
    OR_Node,
    NOR_Node,
    XOR_Node,
    XNOR_Node,
]

# -------------------------------------------


# ARITHMETIC --------------------------------


class ArithmeticNodeBase(OperatorNodeBase):
    color = "#58db53"


class Plus_Node(ArithmeticNodeBase):
    title = "+"

    def apply_op(self, elements: list):
        v = elements[0]
        for e in elements[1:]:
            v = v + e
        return v
        # return sum(elements)


class Minus_Node(ArithmeticNodeBase):
    title = "-"

    def apply_op(self, elements: list):
        v = elements[0]
        for e in elements[1:]:
            v = v - e
        return v
        # return sum(elements[:1])-sum(elements[1:])


class Multiply_Node(ArithmeticNodeBase):
    title = "*"

    def apply_op(self, elements: list):
        v = elements[0]
        for e in elements[1:]:
            v *= e
        return v
        # return math.prod(elements)


class Divide_Node(ArithmeticNodeBase):
    title = "/"

    def apply_op(self, elements: list):
        v = elements[0]
        for e in elements[1:]:
            v = v / e
        return v
        # if len(elements) > 0:
        #     x = elements[0]
        #     for e in elements[1:]:
        #         x /= e
        #     return x
        # else:
        #     return None


class Power_Node(ArithmeticNodeBase):
    title = "**"

    def apply_op(self, elements: list):
        v = elements[0]
        for e in elements[1:]:
            v = v**e
        return v
        # if len(elements) > 0:
        #     x = elements[0]
        #     for e in elements[1:]:
        #         x **= e
        #     return x
        # else:
        #     return None


arithmetic_nodes = [
    Plus_Node,
    Minus_Node,
    Multiply_Node,
    Divide_Node,
    Power_Node,
]

# -------------------------------------------


# COMPARATORS -------------------------------


class ComparatorNodeBase(OperatorNodeBase):
    color = "#a1574c"

    def apply_op(self, elements: list):
        # if len(elements) > 0:
        b = True
        for i in range(1, len(elements)):
            b = b and (self.comp(elements[i - 1], elements[i]))
        return b
        # return None

    def comp(self, a, b) -> bool:
        return False


class Equal_Node(ComparatorNodeBase):
    title = "=="

    def comp(self, a, b) -> bool:
        return a == b


class NotEqual_Node(ComparatorNodeBase):
    title = "!="

    def comp(self, a, b) -> bool:
        return a != b


class Greater_Node(ComparatorNodeBase):
    title = ">"

    def comp(self, a, b) -> bool:
        return a > b


class GreaterEq_Node(ComparatorNodeBase):
    title = ">="

    def comp(self, a, b) -> bool:
        return a >= b


class Less_Node(ComparatorNodeBase):
    title = "<"

    def comp(self, a, b) -> bool:
        return a < b


class LessEq_Node(ComparatorNodeBase):
    title = "<="

    def comp(self, a, b) -> bool:
        return a <= b


comparator_nodes = [
    Equal_Node,
    NotEqual_Node,
    Greater_Node,
    GreaterEq_Node,
    Less_Node,
    LessEq_Node,
]

# -------------------------------------------


nodes = [
    *logic_nodes,
    *arithmetic_nodes,
    *comparator_nodes,
]



import unittest
import ironflow


class TestVersion(unittest.TestCase):
    def test_version(self):
        version = ironflow.__version__
        print(version)
        self.assertTrue(version.startswith('0'))



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
A wrapper for `ryvencore.Session` that makes sure our custom flow class gets
instantiated
"""

from __future__ import annotations

from typing import Optional

from ryvencore.Session import Session as SessionCore

from ironflow.model.script import Script


class Session(SessionCore):
    """
    Wraps `ryvencore.Session.Session` to create our `Script` class instead
    """

    def create_script(
        self,
        title: Optional[str] = None,
        create_default_logs: bool = True,
        data: Optional[dict] = None,
    ) -> Script:
        """
        Creates and returns a new script.
        If data is provided the title parameter will be ignored.
        """
        # Exactly the content of ScriptCore.create_script,
        # but the `Script` class here is different!

        script = Script(
            session=self,
            title=title,
            create_default_logs=create_default_logs,
            load_data=data,
        )

        self.scripts.append(script)
        script.load_flow()

        self.new_script_created.emit(script)

        return script



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Abstract class for high-level GUI objects
"""

from abc import ABC, abstractmethod

from ipywidgets import Box


class Screen(ABC):
    """
    Abstract class for high-level GUI objects
    """

    # Data methods

    @property
    @abstractmethod
    def screen(self) -> Box:
        pass

    def draw(self):
        return self.screen

    # UI methods



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
An interface for doing the visual scripting
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import ipywidgets as widgets

from ironflow.gui.draws_widgets import DrawsWidgets, draws_widgets
from ironflow.gui.workflows.boxes.flow import FlowBox
from ironflow.gui.workflows.boxes.node_interface.control import NodeController
from ironflow.gui.workflows.boxes.node_interface.representation import NodePresenter
from ironflow.gui.workflows.boxes.text_output import TextOut
from ironflow.gui.workflows.boxes.toolbar import Toolbar
from ironflow.gui.workflows.boxes.user_input import UserInput
from ironflow.gui.workflows.canvas_widgets.flow import FlowCanvas
from ironflow.utils import display_string

if TYPE_CHECKING:
    from ironflow.gui.gui import GUI
    from ironflow.gui.workflows.canvas_widgets.nodes import NodeWidget
    from ironflow.model.flow import Flow
    from ironflow.model.node import Node
    from ironflow.model.port import NodeInput, NodeOutput


class WorkflowsGUI(DrawsWidgets):
    main_widget_class = widgets.VBox

    def __new__(cls, gui, *args, **kwargs):
        return super().__new__(cls, *args, **kwargs)

    def __init__(self, gui: GUI, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.gui = gui
        self.flow_canvases = []

        self.toolbar = Toolbar()
        self.node_controller = NodeController(self)
        self.node_presenter = NodePresenter()
        self.text_out = TextOut()
        self.input = UserInput()
        self.flow_box = FlowBox(self.gui.nodes_dictionary)

        self.toolbar.alg_mode_dropdown.observe(
            self._change_alg_mode_dropdown, names="value"
        )
        self.toolbar.buttons.help_node.on_click(self._click_node_help)
        self.toolbar.buttons.load.on_click(self._click_load)
        self.toolbar.buttons.save.on_click(self._click_save)
        self.toolbar.buttons.add_node.on_click(self._click_add_node)
        self.toolbar.buttons.delete_node.on_click(self._click_delete_node)
        self.toolbar.buttons.create_script.on_click(self._click_create_script)
        self.toolbar.buttons.rename_script.on_click(self._click_rename_script)
        self.toolbar.buttons.delete_script.on_click(self._click_delete_script)
        self.toolbar.buttons.zero_location.on_click(self._click_zero_location)
        self.toolbar.buttons.zoom_in.on_click(self._click_zoom_in)
        self.toolbar.buttons.zoom_out.on_click(self._click_zoom_out)
        self.flow_box.script_tabs.observe(self._change_script_tabs)

        self.widget.children = [
            self.toolbar.widget,
            self.input.widget,
            self.flow_box.widget,
            self.text_out.widget,
            widgets.HBox([self.node_controller.widget, self.node_presenter.widget]),
        ]

    @property
    def new_node_class(self):
        return self.flow_box.node_selector.new_node_class

    @property
    def flow_canvas(self):
        return self.flow_canvases[self.gui.active_script_index]

    @property
    def selected_node(self) -> Node | None:
        selected = self.flow_canvas.get_selected_objects()
        return selected[0].node if len(selected) > 0 else None

    def update_tabs(self):
        self.flow_box.update_tabs(
            outputs=[fc.output for fc in self.flow_canvases],
            titles=[fc.title for fc in self.flow_canvases],
            active_index=self.gui.active_script_index,
        )
        for fc in self.flow_canvases:
            fc.display()

    @draws_widgets
    def add_flow(self, flow: Flow):
        self.flow_canvases.append(FlowCanvas(screen=self, flow=flow))

    @draws_widgets
    def load_from_data(self, data: dict):
        self.flow_canvases = []
        for i_script, script in enumerate(self.gui.session.scripts):
            flow_canvas = FlowCanvas(screen=self, flow=script.flow)
            all_data = data["scripts"][i_script]["flow"]["nodes"]
            for i_node, node in enumerate(script.flow.nodes):
                flow_canvas.load_node(
                    all_data[i_node]["pos x"], all_data[i_node]["pos y"], node
                )
            flow_canvas._built_object_to_gui_dict()
            flow_canvas.redraw()
            self.flow_canvases.append(flow_canvas)

    def open_node_control(self, node: Node) -> None:
        self.node_controller.draw_for_node(node)

    def update_node_control(self) -> None:
        self.node_controller.update()

    def close_node_control(self) -> None:
        self.node_controller.clear()

    def ensure_node_not_controlled(self, node: Node) -> None:
        if self.node_controller.node == node:
            self.node_controller.clear()

    def open_node_presenter(self, node_widget: NodeWidget):
        self.node_presenter.draw_for_node_widget(node_widget)

    def update_node_presenter(self):
        self.node_presenter.update()

    def close_node_presenter(self):
        self.node_presenter.clear()

    def ensure_node_not_presented(self, node_widget: NodeWidget) -> None:
        if self.node_presenter.node_widget == node_widget:
            self.node_presenter.clear()

    def redraw_active_flow_canvas(self):
        self.flow_canvas.redraw()

    def print(self, msg: str):
        self.text_out.print(msg)

    def update_nodes_selector(self, nodes_dictionary: dict):
        self.flow_box.node_selector.update(nodes_dictionary)

    def delete_flow(self, i: int):
        self.flow_canvases.pop(i)
        self.node_controller.close()
        self.node_presenter.clear()

    def _change_alg_mode_dropdown(self, change: dict) -> None:
        # Current behaviour: Updates the flow mode for all scripts
        # Todo: Change only for the active script, and update the dropdown on tab (script) switching
        for script in self.gui.session.scripts:
            script.flow.set_algorithm_mode(self.toolbar.alg_mode_dropdown.value)

    def _click_save(self, change: dict) -> None:
        self.input.open_text(
            "Save file",
            self._click_confirm_save,
            self.gui.session_title,
            description_tooltip="Save to file name (omit the file extension, .json)",
        )
        self.print("Choose a file name to save to (omit the file extension, .json)")

    def _click_confirm_save(self, change: dict) -> None:
        file_name = self.input.text
        self.gui.save(f"{file_name}.json")
        self.print(f"Session saved to {file_name}.json")
        self.input.clear()

    def _click_load(self, change: dict) -> None:
        self.input.open_text(
            "Load file",
            self._click_confirm_load,
            self.gui.session_title,
            description_tooltip="Load from file name (omit the file extension, .json).",
        )
        self.print("Choose a file name to load (omit the file extension, .json)")

    def _click_confirm_load(self, change: dict) -> None:
        file_name = self.input.text
        self.gui.load(f"{file_name}.json")
        self.update_tabs()
        self.node_presenter.clear()
        self.print(f"Session loaded from {file_name}.json")
        self.input.clear()

    def _click_node_help(self, change: dict) -> None:
        self.print(
            display_string(
                f"{self.new_node_class.__name__.replace('_Node', '')}:\n{self.new_node_class.__doc__}"
            )
        )

    def _click_add_node(self, change: dict) -> None:
        self.flow_canvas.add_node(10, 10, self.new_node_class)

    def _click_delete_node(self, change: dict) -> None:
        self.flow_canvas.delete_selected()

    def _click_create_script(self, change: dict) -> None:
        self.gui.create_script()
        self.update_tabs()

    def _click_rename_script(self, change: dict) -> None:
        self.input.open_text(
            "New name",
            self._click_confirm_rename,
            self.gui.script.title,
            description_tooltip="New script name",
        )
        self.print("Choose a new name for the current script")

    def _click_confirm_rename(self, change: dict) -> None:
        new_name = self.input.text
        old_name = self.gui.script.title
        rename_success = self.gui.rename_script(new_name)
        if rename_success:
            self.flow_box.script_tabs.set_title(self.gui.active_script_index, new_name)
            self.print(f"Script '{old_name}' renamed '{new_name}'")
        else:
            self.print(
                f"INVALID NAME: Failed to rename script '{self.gui.script.title}' to '{new_name}'."
            )

    def _click_delete_script(self, change: dict) -> None:
        self.input.open_bool(
            f"Delete the entire script {self.gui.script.title}?",
            self._click_confirm_delete_script,
        )

    def _click_confirm_delete_script(self, change: dict) -> None:
        script_name = self.gui.script.title
        self.delete_flow(self.gui.active_script_index)
        self.gui.delete_script()
        self.update_tabs()
        self.print(f"Script {script_name} deleted")

    def _click_zero_location(self, change: dict) -> None:
        self.flow_canvas.x = 0
        self.flow_canvas.y = 0
        self.flow_canvas.redraw()

    def _click_zoom_in(self, change: dict) -> None:
        self.flow_canvas.zoom_in()

    def _click_zoom_out(self, change: dict) -> None:
        self.flow_canvas.zoom_out()

    def _click_input_text_cancel(self, change: dict) -> None:
        self.input.clear()
        self.text_out.clear()

    def _change_script_tabs(self, change: dict):
        if change["name"] == "selected_index" and change["new"] is not None:
            self.input.clear()
            self.flow_canvas.deselect_all()
            if self.flow_box.script_tabs.selected_index == self.gui.n_scripts:
                self.gui.create_script()
                self.update_tabs()
            else:
                self.gui.active_script_index = self.flow_box.script_tabs.selected_index

    def close(self):
        self.toolbar.close()
        self.node_controller.close()
        self.node_presenter.close()
        self.text_out.close()
        self.input.close()
        self.flow_box.close()
        super().close()



from __future__ import annotations

from ironflow.model import dtypes
from ironflow.model.node import Node
from ironflow.model.port import NodeInputBP, NodeOutputBP
from ironflow.node_tools import main_widgets
from ironflow.nodes.deprecated.special_nodes import DualNodeBase


class ForEach_Node(Node):
    title = "ForEach"
    version = "v0.1"
    init_inputs = [
        NodeInputBP(type_="exec", label="start"),
        NodeInputBP(type_="exec", label="reset"),
        NodeInputBP(dtype=dtypes.List(), label="elements"),
    ]
    init_outputs = [
        NodeOutputBP(label="loop", type_="exec"),
        NodeOutputBP(label="e", type_="data"),
        NodeOutputBP(label="finished", type_="exec"),
    ]
    color = "#b33a27"

    _count = 0

    def update_event(self, inp=-1):
        if inp == 0:
            self._count += 1
            if len(self.inputs.values.elements) > self._count:
                e = self.inputs.values.elements[self._count]
                self.set_output_val(1, e)
                self.exec_output(0)
            else:
                self.exec_output(2)
        elif inp > 0:
            self._count = 0
        self.val = self._count


class ExecCounter_Node(DualNodeBase):
    title = "ExecCounter"
    version = "v0.1"
    init_inputs = [
        NodeInputBP(type_="exec"),
    ]
    init_outputs = [
        NodeOutputBP(type_="exec"),
    ]
    color = "#5d95de"

    def __init__(self, params):
        super().__init__(params, active=True)
        self._count = 0

    def update_event(self, inp=-1):
        if self.active and inp == 0:
            self._count += 1
            self.val = self._count
        elif not self.active:
            self.val = self.input(0)


class Click_Node(Node):
    title = "Click"
    version = "v0.1"
    main_widget_class = main_widgets.ButtonNodeWidget
    init_inputs = []
    init_outputs = [NodeOutputBP(type_="exec")]
    color = "#99dd55"

    def update_event(self, inp=-1):
        self.exec_output(0)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
For richer representations of node data allowing deeper analysis.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import ipywidgets as widgets
from IPython.display import display

from ironflow.gui.draws_widgets import DrawsWidgets, draws_widgets

if TYPE_CHECKING:
    from ironflow.gui.workflows.canvas_widgets import NodeWidget


class NodePresenter(DrawsWidgets):
    """Handles the display of nodes with a representation."""

    main_widget_class = widgets.VBox

    def __init__(self):
        super().__init__()
        self.node_widget = None
        self._widgets = []
        self._toggles = []

        self._toggle_box = widgets.HBox([], layout={"flex_flow": "row wrap"})
        self._representation_box = widgets.VBox([], layout={"max_height": "325px"})

        self._border = "1px solid black"
        self.widget.children = [self._toggle_box, self._representation_box]
        self.widget.layout.width = "100%"
        self.widget.layout.border = ""

    def draw_for_node_widget(self, node_widget: NodeWidget):
        self.clear()
        self.node_widget = node_widget
        if node_widget is not None:
            self.draw()

    def update(self) -> None:
        if (
            self.node_widget is not None
            and self.node_widget.node.representation_updated
        ):
            self.draw()

    @draws_widgets
    def draw(self):
        representations_dict = self.node_widget.node.representations

        if len(representations_dict) != len(self._widgets):
            self._widgets = self._build_widgets(representations_dict)
            self._toggles = self._build_toggles(representations_dict)

        for toggle, widget, representation in zip(
            self._toggles, self._widgets, representations_dict.values()
        ):
            widget.clear_output()
            widget.layout.border = ""
            if toggle.value:
                widget.layout.border = self._border
                with widget:
                    display(representation)

        self._toggle_box.children = self._toggles
        self._representation_box.children = self._widgets

        self.node_widget.node.representation_updated = False
        return self.widget

    @staticmethod
    def _build_widgets(representations: dict) -> list[widgets.Output]:
        return [
            widgets.Output(layout={"border": "solid 1px gray"}) for _ in representations
        ]

    def _build_toggles(self, representations: dict) -> list[widgets.Checkbox]:
        toggles = []
        for i, label in enumerate(representations.keys()):
            toggle = widgets.Checkbox(
                description=label, value=i == 0, indent=False, layout={"width": "100px"}
            )
            toggle.observe(self._on_toggle)
            toggles.append(toggle)
        return toggles

    def _on_toggle(self, change: dict) -> None:
        if change["name"] == "value":
            self.draw()

    def clear(self):
        if self.node_widget is not None:
            self.node_widget.represent_button.pressed = False
            self.node_widget.represent_button.draw()  # Re-draw it as un-pressed
        self.node_widget = None

        self._widgets = []
        self._toggles = []

        self.widget.layout.border = ""
        self._toggle_box.children = []
        self._representation_box.children = []
        super().clear()



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Wrappers and extensions of the underlying ryven model.

The `model.HasSession` class provides the main interface between ironflow and ryven.
The rest of these modules revolve mostly around the `DType` class -- namely, ensuring
that each port has a `dtype` attribute (`None` values are handled fine to describe
un-typed ports) and that these dtypes can be used for meaningful type checking when
making connections.
"""

from ryvencore.NodePort import NodePort



from __future__ import annotations

import pickle
from io import BytesIO

import seaborn as sns
from matplotlib import pylab as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from nglview import NGLWidget

from ironflow.model import dtypes
from ironflow.model.node import Node
from ironflow.model.port import NodeInputBP, NodeOutputBP
from pyiron_atomistics import Atoms


class Plot3d_Node(Node):
    """
    Plot a structure with NGLView.

    Inputs:
        structure (pyiron_atomistics.Atoms): The structure to plot.

    Outputs:
        plot3d (nglview.widget.NGLWidget): The plot object.
        structure (pyiron_atomistics.Atoms): The raw structure object passed in.
    """

    title = "Plot3d"
    version = "v0.1"
    init_inputs = [
        NodeInputBP(dtype=dtypes.Data(valid_classes=Atoms), label="structure"),
    ]
    init_outputs = [
        NodeOutputBP(dtype=dtypes.Data(valid_classes=NGLWidget), label="plot3d"),
        NodeOutputBP(dtype=dtypes.Data(valid_classes=Atoms), label="structure"),
    ]
    color = "#5d95de"

    def update_event(self, inp=-1):
        self.set_output_val(0, self.inputs.values.structure.plot3d())
        self.set_output_val(1, self.inputs.values.structure)


class Matplot_Node(Node):
    """
    A 2D matplotlib plot.

    Inputs:
        x (list | numpy.ndarray): Data for the x-axis.
        y (list | numpy.ndarray): Data for the y-axis.
        fig (Figure | None): The figure to plot to.
        marker (matplotlib marker choice | None): Marker style.
        linestyle (matplotlib linestyle choice | None): Line style.
        color (str): HTML or hex color name.
        alpha (float): Transparency.
        label (str | None): Legend.
        xlabel (str | None): X-axis label.
        ylabel (str | None): Y-axis label.
        title (str | None): Figure title.
        legend (bool): Whether to add the legend.
        tight_layout (bool): Call matplotlib `tight_layout` command.

    Outputs:
        fig (matplotlib.figure.Figure): The resulting figure after a
        `matplotlib.pyplot.plot` call on x and y.
    """

    title = "MatPlot"
    version = "v0.1"
    init_inputs = [
        NodeInputBP(dtype=dtypes.Untyped(), label="x"),
        NodeInputBP(dtype=dtypes.Untyped(), label="y"),
        NodeInputBP(
            dtype=dtypes.Data(valid_classes=Figure, allow_none=True), label="fig"
        ),
        NodeInputBP(
            dtype=dtypes.Choice(
                default="o",
                items=[
                    "none",
                    ".",
                    ",",
                    "o",
                    "v",
                    "^",
                    "<",
                    ">",
                    "1",
                    "2",
                    "3",
                    "4",
                    "8",
                    "s",
                    "p",
                    "P",
                    "*",
                    "h",
                    "H",
                    "+",
                    "x",
                    "X",
                    "d",
                    "D",
                    "|",
                    "_",
                ],
            ),
            label="marker",
        ),
        NodeInputBP(
            dtype=dtypes.Choice(
                default="none",
                items=["none", "solid", "dotted", "dashed", "dashdot"],
            ),
            label="linestyle",
        ),
        NodeInputBP(dtype=dtypes.String(default=None, allow_none=True), label="color"),
        NodeInputBP(dtype=dtypes.Float(default=1.0, bounds=(0.0, 1.0)), label="alpha"),
        NodeInputBP(dtype=dtypes.String(default=None, allow_none=True), label="label"),
        NodeInputBP(dtype=dtypes.String(default=None, allow_none=True), label="xlabel"),
        NodeInputBP(dtype=dtypes.String(default=None, allow_none=True), label="ylabel"),
        NodeInputBP(dtype=dtypes.String(default=None, allow_none=True), label="title"),
        NodeInputBP(dtype=dtypes.Boolean(default=False), label="legend"),
        NodeInputBP(dtype=dtypes.Boolean(default=True), label="tight_layout"),
    ]
    init_outputs = [
        NodeOutputBP(dtype=dtypes.Data(valid_classes=Figure), label="fig"),
    ]
    color = "#5d95de"

    def update_event(self, inp=-1):
        super().update_event()
        plt.ioff()
        if self.all_input_is_valid:
            try:
                if self.inputs.values.fig is None:
                    fig, ax = plt.subplots()
                else:
                    fig, ax = self.deepcopy_matplot(self.inputs.values.fig)
                ax.plot(
                    self.inputs.values.x,
                    self.inputs.values.y,
                    marker=self.inputs.values.marker,
                    linestyle=self.inputs.values.linestyle,
                    color=self.inputs.values.color,
                    alpha=self.inputs.values.alpha,
                    label=self.inputs.values.label,
                )
                if self.inputs.values.xlabel is not None:
                    ax.set_xlabel(self.inputs.values.xlabel)
                if self.inputs.values.ylabel is not None:
                    ax.set_ylabel(self.inputs.values.ylabel)
                if self.inputs.values.title is not None:
                    ax.set_title(self.inputs.values.title)
                if self.inputs.values.legend:
                    fig.legend()
                if self.inputs.values.tight_layout:
                    fig.tight_layout()
                self.set_output_val(0, fig)
                plt.ion()
            except Exception as e:
                self.set_all_outputs_to_none()
                plt.ion()
                raise e

    @staticmethod
    def deepcopy_matplot(fig: Figure) -> tuple[Figure, Axes]:
        # Courtesty of StackOverflow @ImportanceOfBeingErnest
        # https://stackoverflow.com/questions/45810557/pyplot-copy-an-axes-content-and-show-it-in-a-new-figure
        buf = BytesIO()
        pickle.dump(fig, buf)
        buf.seek(0)
        fig_copy = pickle.load(buf)
        return fig_copy, fig_copy.axes[0]


_seaborn_method_map = {
    "scatter": sns.scatterplot,
    "hist": sns.histplot,
    "joint": sns.jointplot,
}


class QuickPlot_Node(Node):
    """
    Make a variety of quick and dirty plots with Seaborn.
    """

    title = "QuickPlot"
    color = "#5d95de"

    init_inputs = [
        NodeInputBP(dtype=dtypes.Untyped(), label="x"),
        NodeInputBP(dtype=dtypes.Untyped(), label="y"),
        NodeInputBP(
            dtype=dtypes.Choice(
                default="scatter",
                items=list(_seaborn_method_map.keys()),
            ),
            label="type",
        ),
    ]
    init_outputs = [NodeOutputBP(label="plot")]

    def update_event(self, inp=-1):
        super().update_event()
        plt.ioff()
        if self.all_input_is_valid:
            try:
                plt.clf()
                plot_function = _seaborn_method_map[self.inputs.values.type]
                out = plot_function(x=self.inputs.values.x, y=self.inputs.values.y)
                self.set_output_val(0, out.figure)
                plt.ion()
            except Exception as e:
                self.set_all_outputs_to_none()
                plt.ion()
                raise e



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Canvas representations of the nodes.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import numpy as np
from ipycanvas import hold_canvas

from ironflow.gui.workflows.canvas_widgets.base import CanvasWidget
from ironflow.gui.workflows.canvas_widgets.buttons import (
    RepresentButtonWidget,
    ExpandButtonWidget,
    CollapseButtonWidget,
    ExecButtonWidget,
)
from ironflow.gui.workflows.canvas_widgets.layouts import (
    NodeLayout,
    DataPortLayout,
    ExecPortLayout,
    ButtonLayout,
)
from ironflow.gui.workflows.canvas_widgets.ports import PortWidget

if TYPE_CHECKING:
    from ironflow.gui.workflows.canvas_widgets.flow import FlowCanvas
    from ironflow.gui.workflows.canvas_widgets.base import Number
    from ironflow.model.port import NodeInputBP, NodeOutputBP
    from ironflow.model.node import Node


class NodeWidget(CanvasWidget):
    """
    The main ipycanvas representation of a node. Has graphical elements for IO ports. Collapsable to save space.

    Also has a `SHOW` button that sends a representation over to the `ironflow.GUI.node_presenter` window.
    Presentation gets locked until the button is pressed again, the node is deleted, or another node gets presented.
    While presented, representation updates automatically on changes to input.
    """

    def __init__(
        self,
        x: Number,
        y: Number,
        parent: FlowCanvas | CanvasWidget,
        layout: NodeLayout,
        node: Node,
        selected: bool = False,
        title: Optional[str] = None,
        port_radius: Number = 10,
    ):
        super().__init__(
            x=x,
            y=y,
            parent=parent,
            layout=layout,
            selected=selected,
            title=title if title is not None else node.title,
        )

        self.node = node
        self.inputs = node.inputs
        self.outputs = node.outputs

        # Register callback to change color on updates
        self.node.widget = self
        self._updating = False
        self.node.before_update.connect(self._draw_before_updating)
        self.node.after_update.connect(self._draw_after_updating)

        self.port_radius = port_radius
        self.port_layouts = {"data": DataPortLayout(), "exec": ExecPortLayout()}

        n_ports_max = (
            max(len(self.node.inputs), len(self.node.outputs)) + 1
        )  # Includes the expand/collapse button
        exec_port_i = np.where([p.type_ == "exec" for p in self.node.inputs])[0]
        n_ports_min = exec_port_i[-1] + 1 if len(exec_port_i) > 0 else 1
        subwidget_size_and_buffer = 1.33 * 2 * self.port_radius
        self._title_box_height = self.layout.title_box_height
        self._max_body_height = subwidget_size_and_buffer * n_ports_max
        self._min_body_height = subwidget_size_and_buffer * n_ports_min
        self._expanded_height = self._title_box_height + self._max_body_height
        self._collapsed_height = self._title_box_height + self._min_body_height
        self._height = self._expanded_height

        y_step = self._max_body_height / n_ports_max
        self._subwidget_y_locs = (
            np.arange(n_ports_max) + 0.5
        ) * y_step + self._title_box_height

        self.add_inputs()
        self.add_outputs()
        self.expand_button = ExpandButtonWidget(
            x=0.5 * self.width - self.port_radius,
            y=self._subwidget_y_locs[0] - self.port_radius,
            parent=self,
            layout=ButtonLayout(),
            pressed=True,
            visible=False,
            size=2 * self.port_radius,
        )
        self.add_widget(self.expand_button)
        self.collapse_button = CollapseButtonWidget(
            x=0.5 * self.width - self.port_radius,
            y=self._subwidget_y_locs[-1] - self.port_radius,
            parent=self,
            layout=ButtonLayout(),
            pressed=False,
            visible=True,
            size=2 * self.port_radius,
        )
        self.add_widget(self.collapse_button)

        button_layout = ButtonLayout()
        button_edge_offset = 5
        self.represent_button = RepresentButtonWidget(
            x=self.width - button_layout.width - button_edge_offset,
            y=button_edge_offset,
            parent=self,
            layout=button_layout,
        )
        self.add_widget(self.represent_button)

    def on_click(
        self, last_selected_object: Optional[CanvasWidget]
    ) -> NodeWidget | None:
        if last_selected_object == self:
            return self
        else:
            if last_selected_object is not None:
                last_selected_object.deselect()
            self.select()
            try:
                self.screen.open_node_control(self.node)
                return self
            except Exception as e:
                self.screen.print(
                    f"Failed to handle selection of {self} with exception {e}"
                )
                self.screen.close_node_control()
                self.deselect()
                return None

    def on_double_click(self) -> None:
        self.delete()
        return None

    @staticmethod
    def _draw_before_updating(node: Node, inp: int) -> None:
        node.widget._updating = True
        with hold_canvas(node.widget.canvas):
            node.widget.draw()

    @staticmethod
    def _draw_after_updating(node: Node, inp: int) -> None:
        node.widget._updating = False
        with hold_canvas(node.widget.canvas):
            node.widget.draw()

    @property
    def color(self) -> str:
        if self._updating:
            return self.layout.updating_color
        elif self.selected:
            return self.layout.selected_color
        else:
            return self.layout.background_color

    def draw_title(self) -> None:
        self.canvas.fill_style = self.node.color
        self.canvas.fill_rect(self.x, self.y, self.width, self._title_box_height)
        self.canvas.font = self.layout.font_string
        self.canvas.fill_style = self.layout.font_color
        x = self.x + (self.width * 0.04)
        y = self.y + self._title_box_height - 8
        self.canvas.fill_text(self.title[: self.layout.max_title_chars], x, y)

    def _add_ports(
        self,
        radius: Number,
        inputs: Optional[list[NodeInputBP]] = None,
        outputs: Optional[list[NodeOutputBP]] = None,
        border: Number = 1.4,
    ) -> None:
        if inputs is not None:
            x = radius * border
            data = inputs
            title_alignment = "start"
        elif outputs is not None:
            x = self.width - radius * border
            data = outputs
            title_alignment = "end"
        else:
            return

        n_ports = len(data)
        for i_port in range(n_ports):
            port = data[i_port]
            data_or_exec = port.type_
            self.add_widget(
                PortWidget(
                    x=x,
                    y=self._subwidget_y_locs[i_port],
                    parent=self,
                    layout=self.port_layouts[data_or_exec],
                    port=port,
                    hidden_x=x,
                    hidden_y=self._subwidget_y_locs[0],
                    radius=radius,
                    title_alignment=title_alignment,
                )
            )
            if data_or_exec == "exec" and inputs is not None:
                button_layout = ButtonLayout()
                self.add_widget(
                    ExecButtonWidget(
                        x=x + radius,
                        y=self._subwidget_y_locs[i_port] - 0.5 * button_layout.height,
                        parent=self,
                        layout=button_layout,
                        port=port,
                    )
                )

    def add_inputs(self) -> None:
        self._add_ports(radius=self.port_radius, inputs=self.inputs)

    def add_outputs(self) -> None:
        self._add_ports(radius=self.port_radius, outputs=self.outputs)

    def delete(self) -> None:
        self.screen.ensure_node_not_presented(self)
        self.screen.ensure_node_not_controlled(self.node)
        for c in self.flow.connections[
            ::-1
        ]:  # Reverse to make sure we traverse whole thing even if we delete
            # Todo: Can we be more efficient than looping over all nodes?
            if (c.inp.node == self.node) or (c.out.node == self.node):
                self.flow.remove_connection(c)
        self.flow.remove_node(self.node)
        self.parent.objects_to_draw.remove(self)

    def deselect(self) -> None:
        super().deselect()
        self.screen.ensure_node_not_controlled(self.node)

    @property
    def port_widgets(self) -> list[PortWidget]:
        return [o for o in self.objects_to_draw if isinstance(o, PortWidget)]

    def expand_io(self):
        self._height = self._expanded_height
        for o in self.port_widgets:
            o.show()
        self.collapse_button.unpress()

    def collapse_io(self):
        self._height = self._collapsed_height
        for o in self.port_widgets:
            o.hide()
        self.expand_button.unpress()


class ButtonNodeWidget(NodeWidget):
    def __init__(
        self,
        x: Number,
        y: Number,
        parent: FlowCanvas | CanvasWidget,
        layout: NodeLayout,
        node: Node,
        selected: bool = False,
        title: Optional[str] = None,
        port_radius: Number = 10,
    ):
        super().__init__(
            x=x,
            y=y,
            parent=parent,
            layout=layout,
            node=node,
            selected=selected,
            title=title,
            port_radius=port_radius,
        )

        button_layout = ButtonLayout()
        self.exec_button = ExecButtonWidget(
            x=0.8 * (self.width - button_layout.width),
            y=self._subwidget_y_locs[0] - 0.5 * button_layout.height,
            parent=self,
            layout=button_layout,
            port=self.node.outputs[0],
        )
        self.add_widget(self.exec_button)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Canvas widgets for ryven IO ports.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Optional

from ironflow.gui.workflows.canvas_widgets.base import HideableWidget, CanvasWidget
from ironflow.gui.workflows.canvas_widgets.layouts import PortLayout

if TYPE_CHECKING:
    from ironflow.gui.workflows.canvas_widgets.flow import FlowCanvas
    from ironflow.gui.workflows.canvas_widgets.base import Number
    from ironflow.model.port import NodeInput, NodeOutput


class PortWidget(HideableWidget):
    def __init__(
        self,
        x: Number,
        y: Number,
        parent: FlowCanvas | CanvasWidget,
        layout: PortLayout,
        port: NodeInput | NodeOutput,
        selected: bool = False,
        title: Optional[str] = None,
        hidden_x: Optional[Number] = None,
        hidden_y: Optional[Number] = None,
        radius: Number = 10,
        title_alignment: Literal["start", "end"] = "start",
    ):
        super().__init__(
            x=x,
            y=y,
            parent=parent,
            layout=layout,
            selected=selected,
            title=title if title is not None else port.label_str,
            hidden_x=hidden_x,
            hidden_y=hidden_y,
        )

        self.radius = radius
        self.port = port
        self.title_alignment = title_alignment
        self.highlighted = False

    def on_click(
        self, last_selected_object: Optional[CanvasWidget]
    ) -> PortWidget | None:
        if last_selected_object == self:
            self.deselect()
            return None
        elif isinstance(last_selected_object, PortWidget):
            self.flow.connect_nodes(last_selected_object.port, self.port)
            self.deselect_all()
            return None
        else:
            if last_selected_object is not None:
                last_selected_object.deselect()
            self.select()
            return self

    @property
    def _current_color(self):
        if self.highlighted:
            color = self.layout.highlight_color
        elif self.port.ready:
            if self.selected:
                color = self.layout.valid_selected_color
            else:
                color = self.layout.valid_color
        else:
            if self.selected:
                color = self.layout.invalid_selected_color
            else:
                color = self.layout.invalid_color
        return color

    def draw_shape(self) -> None:
        self.canvas.fill_style = self._current_color
        self.canvas.fill_circle(self.x, self.y, self.radius)

    def draw_title(self) -> None:
        self.canvas.font = self.layout.font_string
        self.canvas.fill_style = self.layout.font_color
        shift_magnitude = self.radius + 3
        if self.title_alignment == "start":
            shift = shift_magnitude
        elif self.title_alignment == "end":
            shift = -shift_magnitude
        else:
            raise ValueError(
                f"Title alignment {self.title_alignment} not recognized, please choose start or end"
            )
        self.canvas.text_align = self.title_alignment
        title = self.title.upper() if self.port.dtype.batched else self.title
        self.canvas.fill_text(
            title[: self.layout.max_title_chars],
            self.x + shift,
            self.y + self.radius // 2,
        )
        self.canvas.text_align = "start"  # Revert to default after writing

    def _is_at_xy(self, x_in: Number, y_in: Number) -> bool:
        return (x_in - self.x) ** 2 + (y_in - self.y) ** 2 < self.radius**2

    def select(self) -> None:
        super().select()
        self.gui.build_recommendations(self.port)
        self.flow_canvas.highlight_compatible_ports(self)

    def deselect(self) -> None:
        super().deselect()
        self.flow_canvas.clear_port_highlighting()

    def highlight(self):
        self.highlighted = True

    def dehighlight(self):
        self.highlighted = False



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Provides lazy-loading of `pyiron_ontology` ontologies for de-serializing otypes.

TODO: This is not easily extensible; it needs a system for users to register their
      ontology with the session so that they can use their own otypes alongside
      modules of their own custom nodes.
"""


class OTypeLoader:
    _atomistics = None

    @classmethod
    def atomistics(cls, item):
        # Note: public attribute name must match saved ontology namespace
        if cls._atomistics is None:
            from pyiron_ontology import AtomisticsOntology

            cls._atomistics = AtomisticsOntology().onto
        return cls._atomistics[item]


def otype_from_str(namespace, item):
    return getattr(OTypeLoader, namespace)(item)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
A new output port class that has a dtype.
"""

from __future__ import annotations

from copy import deepcopy
from typing import Optional, TYPE_CHECKING

from numpy import argwhere
from ryvencore.InfoMsgs import InfoMsgs
from ryvencore.NodePort import NodeInput as NodeInputCore, NodeOutput as NodeOutputCore
from ryvencore.NodePortBP import (
    NodeOutputBP as NodeOutputBPCore,
    NodeInputBP as NodeInputBPCore,
)
from ryvencore.utils import serialize

from ironflow.model.dtypes import DType, Untyped

if TYPE_CHECKING:
    from ironflow.model.node import Node


class HasDType:
    """A mixin to add the valid value check property"""

    def set_dtype_ok(self):
        if self.dtype is not None:
            if self.val is not None:
                self._dtype_ok = self.dtype.valid_val(self.val)
            else:
                self._dtype_ok = self.dtype.allow_none
        else:
            self._dtype_ok = True

    @property
    def dtype_ok(self):
        try:
            return self._dtype_ok
        except AttributeError:
            self.set_dtype_ok()
            return self._dtype_ok


class HasOType:
    """A mixin to add the valid value check to properties with an ontology type"""

    def recalculate_otype_checks(self, ignore=None):
        self.set_otype_ok()
        if self.otype is not None:
            # Along connections
            for con in self.connections:
                if isinstance(self, NodeInput):
                    other = con.out
                else:
                    other = con.inp

                if other != ignore and other.otype is not None:
                    other.recalculate_otype_checks(ignore=self)

            # Across the node
            if isinstance(self, NodeInput):
                ports = self.node.outputs.ports
            else:
                ports = self.node.inputs.ports
            if ignore not in ports:
                for port in ports:
                    if port.otype is not None:
                        port.recalculate_otype_checks(ignore=self)

    def set_otype_ok(self):
        if self.otype is not None:
            if isinstance(self, NodeInput):
                input_tree = self.otype.get_source_tree(
                    additional_requirements=self.get_downstream_requirements()
                )
                self._otype_ok = all(
                    con.out.all_connections_found_in(input_tree)
                    for con in self.connections
                    if con.out.otype is not None
                )
            else:
                self._otype_ok = all(
                    con.inp.workflow_tree_contains_connections_of(self)
                    for con in self.connections
                    if con.inp.otype is not None
                )
        else:
            self._otype_ok = True

    @property
    def otype_ok(self):
        try:
            return self._otype_ok
        except AttributeError:
            self.set_otype_ok()
            return self._otype_ok

    def _output_graph_is_represented_in_workflow_tree(self, output_port, input_tree):
        try:
            output_index = argwhere(
                [output_port.otype == source.value for source in input_tree.children]
            )[0][0]
            upstream_inputs = [
                inp
                for inp in output_port.node.inputs
                if inp.otype is not None and len(inp.connections) > 0
            ]
            for usi in upstream_inputs:
                input_branches = input_tree.children[output_index].children[0].children
                # input/generic->outputs->function->inputs

                input_index = argwhere(
                    [usi.otype == source.value for source in input_branches]
                )[0][0]

                for con in usi.connections:
                    if (
                        con.out.otype is not None
                        and not self._output_graph_is_represented_in_workflow_tree(
                            con.out, input_branches[input_index]
                        )
                    ):
                        return False
            return True
        except IndexError:
            # Can't slice argwhere if it finds nothing
            return False

    def get_downstream_requirements(self):
        downstream_requirements = []
        for out in self.node.outputs:
            if out.otype is not None:
                for downstream_inp in [conn.inp for conn in out.connections]:
                    if downstream_inp.otype is not None:
                        downstream_requirements += (
                            downstream_inp.get_downstream_requirements()
                        )
        try:
            return self.otype.get_requirements(list(set(downstream_requirements)))
        except AttributeError:
            return list(set(downstream_requirements))


class HasTypes(HasOType, HasDType):
    @property
    def ready(self):
        return self.dtype_ok and self.otype_ok


class NodeInput(NodeInputCore, HasTypes):
    def __init__(
        self,
        node: Node,
        type_: str = "data",
        label_str: str = "",
        add_data: Optional[dict] = None,
        dtype: Optional[DType] = None,
        otype=None,
    ):
        super().__init__(
            node=node,
            type_=type_,
            label_str=label_str,
            add_data=add_data if add_data is not None else {},
            dtype=Untyped() if dtype is None else deepcopy(dtype),
        )
        self.otype = otype

    def _update_node(self):
        self.node.update(self.node.inputs.index(self))

    def batch(self):
        if self.dtype is not None and not self.dtype.batched:
            self.dtype.batched = True
            if len(self.connections) == 0:
                self.update([self.val])
            else:
                self.set_dtype_ok()
                self._update_node()

    def unbatch(self):
        if self.dtype is not None and self.dtype.batched:
            self.dtype.batched = False
            if len(self.connections) == 0:
                self.update(self.val[-1])
            else:
                self.set_dtype_ok()
                self._update_node()

    def data(self) -> dict:
        data = super().data()

        if self.otype is not None:
            data["otype_namespace"] = self.otype.namespace.name
            data["otype_name"] = self.otype.name

        return data

    def workflow_tree_contains_connections_of(self, port: NodeOutput):
        tree = self.otype.get_source_tree(
            additional_requirements=self.get_downstream_requirements()
        )
        return self._output_graph_is_represented_in_workflow_tree(port, tree)

    def update(self, data=None):
        # super().update(data=data)
        # We need to add the dtype update _between_ the val update and node update
        if self.type_ == "data":
            self.val = data  # self.get_val()
            InfoMsgs.write("Data in input set to", data)

        self.set_dtype_ok()
        self._update_node()

    def connected(self):
        super().connected()
        self.set_dtype_ok()
        self.recalculate_otype_checks()  # Note: Only need to call or one of input or
        # output since Flow.add_connection calls .connected on both inp and out
        if self.dtype is not None or self.otype is not None:
            # Force a re-calculation in case the update had type-matching restrictions
            self._update_node()

    def disconnected(self):
        super().disconnected()
        self.recalculate_otype_checks()


class NodeOutput(NodeOutputCore, HasTypes):
    def __init__(
        self,
        node,
        type_="data",
        label_str="",
        dtype: Optional[DType] = None,
        otype=None,
    ):
        super().__init__(node=node, type_=type_, label_str=label_str)
        self.dtype = Untyped() if dtype is None else deepcopy(dtype)
        self.otype = otype

    def data(self) -> dict:
        data = super().data()

        if self.dtype is not None:
            data["dtype"] = str(self.dtype)
            data["dtype state"] = serialize(self.dtype.get_state())

        if self.otype is not None:
            data["otype_namespace"] = self.otype.namespace.name
            data["otype_name"] = self.otype.name

        return data

    def all_connections_found_in(self, tree):
        """
        Checks to see if actual ontologically typed connections match with all
        ontologically possible workflows for an input port.
        """
        return self._output_graph_is_represented_in_workflow_tree(self, tree)

    def set_val(self, val):
        super().set_val(val)
        self.set_dtype_ok()

    def disconnected(self):
        super().disconnected()
        self.recalculate_otype_checks()
        # Unlike `connected`, we do need to explicitly recalculate on the disconnected
        # output, because it is no longer part of the input's graph tree!


class NodeInputBP(NodeInputBPCore):
    def __init__(
        self,
        label: str = "",
        type_: str = "data",
        dtype: DType = None,
        add_data={},
        otype=None,
    ):
        super().__init__(label=label, type_=type_, dtype=dtype, add_data=add_data)
        self.otype = otype


class NodeOutputBP(NodeOutputBPCore):
    def __init__(
        self,
        label: str = "",
        type_: str = "data",
        dtype: Optional[DType] = None,
        otype=None,
    ):
        super().__init__(label=label, type_=type_)
        self.dtype = dtype
        self.otype = otype



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
The necessary classes for creating new nodes.

Example:
    >>> from ironflow import GUI
    >>> from ironflow.node_tools import Node, NodeInputBP, NodeOutputBP, dtypes, input_widgets
    >>> gui = GUI(script_title='foo')
    >>>
    >>> class MyNode(Node):
    >>>     title = "MyUserNode"
    >>>     init_inputs = [
    >>>         NodeInputBP(dtype=dtypes.Integer(default=1), label="foo")
    >>>     ]
    >>>     init_outputs = [
    >>>        NodeOutputBP(label="bar")
    >>>    ]
    >>>    color = 'cyan'
    >>>
    >>>     def update_event(self, inp=-1):
    >>>         self.set_output_val(0, self.input(0) + 42)
    >>>
    >>> gui.register_node(MyNode)
"""

import ironflow.node_tools.input_widgets
import ironflow.node_tools.main_widgets
from ironflow.model import dtypes
from ironflow.model.node import (
    Node,
    PlaceholderWidgetsContainer,
    PortList,
    BatchingNode,
    DataNode,
    JobMaker,
    JobNode,
    JobTaker,
)
from ironflow.model.port import NodeInputBP, NodeOutputBP



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Shared functionality among all ipywidgets components of the main gui.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Type

import ipywidgets as widgets


class Box(ABC):
    def __init__(self):
        self._box = self.box_class([], layout=self.layout)

    @property
    @abstractmethod
    def box_class(self) -> Type[widgets.Box]:
        """E.g. `widgets.HBox` or `widgets.VBox`"""
        pass

    @property
    def box(self) -> widgets.Box:
        return self._box

    @property
    def layout(self) -> widgets.Layout:
        """An empty layout. Overwrite in children as desired."""
        return widgets.Layout()

    def clear(self):
        self.box.children = []



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
For giving text feedback to the user.
"""

from __future__ import annotations

import ipywidgets as widgets
from IPython.display import display

from ironflow.gui.draws_widgets import DrawsWidgets


class TextOut(DrawsWidgets):
    main_widget_class = widgets.VBox

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.widget.width = "100%"
        self.widget.border = "1px solid black"
        self._output = widgets.Output()
        self._button = widgets.Button(
            tooltip="Clear output",
            description="clear",
            layout={"width": "100px"},
        )
        self._button.on_click(self._click_button)

    def clear(self):
        self._output.clear_output()
        self.widget.children = []
        super().clear()

    def _click_button(self, change: dict) -> None:
        self.clear()

    def print(self, msg: str):
        self._output.clear_output()
        with self._output:
            display(msg)
        self.widget.children = [self._output, self._button]



from __future__ import annotations

import numpy as np

from ironflow.model import dtypes
from ironflow.model.node import DataNode
from ironflow.model.port import NodeInputBP, NodeOutputBP


class Select_Node(DataNode):
    """
    Select a single elemnt of an iterable input.
    """

    title = "Select"
    init_inputs = [
        NodeInputBP(dtype=dtypes.List(valid_classes=object), label="array"),
        NodeInputBP(dtype=dtypes.Integer(default=0), label="i"),
    ]
    init_outputs = [
        NodeOutputBP(label="item", dtype=dtypes.Data(valid_classes=object)),
    ]
    color = "#aabb44"

    def update_event(self, inp=-1):
        self.outputs.ports.item.dtype.valid_classes = [
            type(self.inputs.values.array[self.inputs.values.i])
        ]
        super().update_event(inp=inp)

    def node_function(self, array, i, **kwargs) -> dict:
        return {"item": array[i]}


class Slice_Node(DataNode):
    """
    Slice a numpy array, list, or tuple, and return it as a numpy array.

    When both `i` and `j` are `None`: Return the input whole.
    When `i` is not `None` and `j` is: Return the slice `[i:]`
    When `i` is `None` and `j` isn't: Return the slice `[:j]`
    When neither are `None`: Return the slice `[i:j]`
    """

    title = "Slice"
    init_inputs = [
        NodeInputBP(dtype=dtypes.List(valid_classes=object), label="array"),
        NodeInputBP(dtype=dtypes.Integer(default=None, allow_none=True), label="i"),
        NodeInputBP(dtype=dtypes.Integer(default=None, allow_none=True), label="j"),
    ]
    init_outputs = [
        NodeOutputBP(label="sliced", dtype=dtypes.List(valid_classes=object)),
    ]
    color = "#aabb44"

    def update_event(self, inp=-1):
        self.outputs.ports.sliced.dtype = dtypes.List(
            valid_classes=list(
                set(
                    type(obj)
                    for obj in self._slice(
                        self.inputs.values.array,
                        self.inputs.values.i,
                        self.inputs.values.j,
                    )
                )
            )
        )
        super().update_event(inp=inp)

    @staticmethod
    def _slice(array, i, j):
        converted = np.array(array)
        if i is None and j is None:
            sliced = converted
        elif i is not None and j is None:
            sliced = converted[i:]
        elif i is None and j is not None:
            sliced = converted[:j]
        else:
            sliced = converted[i:j]
        return sliced

    def node_function(self, array, i, j, **kwargs) -> dict:
        return {"sliced": self._slice(array, i, j)}


class Transpose_Node(DataNode):
    """
    Interprets list-like input as a numpy array and transposes it.
    """

    title = "Transpose"
    init_inputs = [
        NodeInputBP(dtype=dtypes.List(valid_classes=object), label="array"),
    ]
    init_outputs = [
        NodeOutputBP(dtype=dtypes.List(valid_classes=object), label="transposed"),
    ]
    color = "#aabb44"

    def node_function(self, array, **kwargs) -> dict:
        array = np.array(array)  # Ensure array
        if len(array.shape) < 2:
            array = np.array([array])  # Ensure transposable
        return {"transposed": np.array(array).T}


class IntRand_Node(DataNode):
    """
    Generate a random non-negative integer.

    Inputs:
        high (int): Biggest possible integer. (Default is 1).
        length (int): How many random numbers to generate. (Default is 1.)

    Outputs:
        randint (int|numpy.ndarray): The randomly generated value(s).
    """

    # this __doc__ string will be displayed as tooltip in the editor

    title = "IntRandom"
    init_inputs = [
        NodeInputBP(dtype=dtypes.Integer(default=0), label="low"),
        NodeInputBP(dtype=dtypes.Integer(default=1), label="high"),
        NodeInputBP(dtype=dtypes.Integer(default=1), label="length"),
    ]
    init_outputs = [
        NodeOutputBP(dtype=dtypes.List(valid_classes=np.integer), label="randint"),
    ]
    color = "#aabb44"

    def node_function(self, low, high, length, *args, **kwargs) -> dict:
        return {"randint": np.random.randint(low, high=high, size=length)}


class Linspace_Node(DataNode):
    """
    Generate a linear mesh in a given range using `np.linspace`.

    Inputs:
        min (int): The lower bound (inclusive). (Default is 1.)
        max (int): The upper bound (inclusive). (Default is 2.)
        steps (int): How many samples to take inside (min, max). (Default is 10.)

    Outputs:
        linspace (numpy.ndarray): A uniform sampling over the requested range.
    """

    # this __doc__ string will be displayed as tooltip in the editor

    title = "Linspace"
    init_inputs = [
        NodeInputBP(dtype=dtypes.Float(default=1.0), label="min"),
        NodeInputBP(dtype=dtypes.Float(default=2.0), label="max"),
        NodeInputBP(dtype=dtypes.Integer(default=10), label="steps"),
    ]
    init_outputs = [
        NodeOutputBP(dtype=dtypes.List(valid_classes=np.floating), label="linspace")
    ]
    color = "#aabb44"

    def node_function(self, min, max, steps, **kwargs) -> dict:
        return {"linspace": np.linspace(min, max, steps)}



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Combines ryven, ipywidgets/ipycanvas, and pyiron to facilitate visual scripting of pyiron workflows.
"""

import ironflow.node_tools
from ironflow.gui.gui import GUI

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Provide access to the `pyiron_gui`
"""

from __future__ import annotations

import ipywidgets as widgets

from pyiron_atomistics import Project
from pyiron_gui import ProjectBrowser

from ironflow.gui.draws_widgets import DrawsWidgets


class BrowserGUI(DrawsWidgets):
    main_widget_class = widgets.VBox

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # self.widget = self.main_widget_class([])
        self.top_level_project = Project(".")
        self.widget.layout.height = "470px"
        self.project_browser = ProjectBrowser(self.top_level_project, Vbox=self.widget)
        self.project_browser.refresh()



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Node classes for the underlying ryven model
"""



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Top-level objects for getting the front and back end (and various parts of the front end) to talk to each other.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Type

import ipywidgets as widgets

from ironflow.gui.browser import BrowserGUI
from ironflow.gui.draws_widgets import DrawsWidgets
from ironflow.gui.log import LogGUI
from ironflow.gui.workflows.screen import WorkflowsGUI
from ironflow.model.model import HasSession

if TYPE_CHECKING:
    from ironflow.model.node import Node
    from ironflow.model.port import NodeInput, NodeOutput


class GUI(HasSession, DrawsWidgets):
    """
    The main ironflow object, connecting a ryven backend with a jupyter-friendly ipywidgets+ipycanvas frontend.

    Methods:
        draw: Build the ipywidget to interact with.
        register_user_node: Register with ironflow a new node from the current python process.
    """

    main_widget_class = widgets.Tab

    def __new__(
        cls,
        session_title: str,
        *args,
        extra_nodes_packages: Optional[list] = None,
        script_title: Optional[str] = None,
        enable_ryven_log: bool = False,
        log_to_display: bool = True,
        **kwargs,
    ):
        return super().__new__(cls, *args, **kwargs)

    def __init__(
        self,
        session_title: str,
        *args,
        extra_nodes_packages: Optional[list] = None,
        script_title: Optional[str] = None,
        enable_ryven_log: bool = False,
        log_to_display: bool = True,
        **kwargs,
    ):
        """
        Create a new gui instance.

        Args:
            session_title (str): Title of the session to use. Will look for a json file of the same name and try to
                read it. If no such file exists, simply makes a new script instead.
            extra_nodes_packages (list | None): an optional list of nodes to register at instantiation. List items can
                be either a list of `ironflow.model.node.Node` subclasses, a module containing such subclasses, or a .py
                file of a module containing such subclasses. In all cases only those subclasses with the name pattern
                `*_Node` will be registered. (Default is None, don't register any extra nodes.)
            script_title (str|None): Title for an initial script. (Default is None, which generates "script_0" if a
                new script is needed on initialization, i.e. when existing session data cannot be read.)
            enable_ryven_log (bool): Activate Ryven's logging system to catch Ryven actions and node errors. (Default
                is True.)
            log_to_display (bool): Re-route stdout (and node error's captured by the Ryven logger, if activated) to a
                separate output widget. (Default is True.)
        """
        self.log = LogGUI(
            model=self, enable_ryven_log=enable_ryven_log, log_to_display=log_to_display
        )
        # Log screen needs to be instantiated before the rest of the init so we know whether to look at the ryven log
        # as we boot

        super().__init__(
            *args,
            session_title=session_title,
            extra_nodes_packages=extra_nodes_packages,
            enable_ryven_log=enable_ryven_log,
            **kwargs,
        )

        self.workflows = WorkflowsGUI(gui=self)
        self.workflows.flow_box.node_selector.select_module(self._starting_module)
        self.browser = BrowserGUI()

        try:
            self.load(f"{self.session_title}.json")
            print(f"Loaded session data for {self.session_title}")
        except FileNotFoundError:
            print(
                f"No session data found for {self.session_title}, making a new script."
            )
            self.create_script(script_title)
        self.workflows.update_tabs()

        self.widget.children = [
            self.workflows.widget,
            self.browser.widget,
            self.log.widget,
        ]
        self.widget.set_title(0, "Workflows")
        self.widget.set_title(1, "Browser")
        self.widget.set_title(2, "Log")

        self.widget.observe(self._change_screen_tabs)

    def create_script(
        self,
        title: Optional[str] = None,
        create_default_logs: bool = True,
        data: Optional[dict] = None,
    ) -> None:
        super().create_script(
            title=title, create_default_logs=create_default_logs, data=data
        )
        self.workflows.add_flow(self.flow)

    @property
    def selected_node(self) -> Node | None:
        return self.workflows.selected_node

    @property
    def new_node_class(self):
        return self.workflows.new_node_class

    def serialize(self) -> dict:
        data = super().serialize()
        currently_active = self.active_script_index
        for i_script, script in enumerate(self.session.scripts):
            all_data = data["scripts"][i_script]["flow"]["nodes"]
            self.active_script_index = i_script
            for i, node_widget in enumerate(self.workflows.flow_canvas.objects_to_draw):
                all_data[i]["pos x"] = node_widget.x
                all_data[i]["pos y"] = node_widget.y
        self.active_script_index = currently_active
        return data

    def load_from_data(self, data: dict) -> None:
        super().load_from_data(data)
        self.workflows.load_from_data(data)

    def register_node(self, node_class: Type[Node], node_group: Optional[str] = None):
        # Inherited __doc__ still applies just fine, all we do here is update a menu item afterwards.
        super().register_node(node_class=node_class, node_group=node_group)
        try:
            self.workflows.update_nodes_selector(self.nodes_dictionary)
        except AttributeError:
            pass  # It's not defined yet in the super().__init__ call, which is fine

    def log_to_display(self):
        self.log.log_to_display()

    def log_to_stdout(self):
        self.log.log_to_stdout()

    def build_recommendations(self, port: NodeInput | NodeOutput):
        self.recommend_nodes(port)
        self.workflows.flow_box.update_nodes(self.nodes_dictionary)

    def clear_recommendations(self):
        self.clear_recommended_nodes()
        self.workflows.flow_box.update_nodes(self.nodes_dictionary)

    def draw(self):
        return self.widget

    # Type hinting for unused `change` argument in callbacks taken from ipywidgets docs:
    # https://ipywidgets.readthedocs.io/en/latest/examples/Widget%20Events.html#Traitlet-events

    def _change_screen_tabs(self, change: dict):
        if change["name"] == "selected_index" and change["new"] == 1:
            self.browser.project_browser.refresh()

    def close(self):
        self.log.close()
        self.browser.close()
        self.workflows.close()
        super().close()



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from ryvencore import Flow as FlowCore, InfoMsgs
from ryvencore.NodePort import NodePort

from ironflow.model.dtypes import Untyped

if TYPE_CHECKING:
    from ironflow.model.dtypes import DType


class Flow(FlowCore):
    """
    Wraps `ryvencore.Flow.Flow` to override the connection validity check, so we can
    add type checking.
    """

    @staticmethod
    def batched_or_nothing(dtype: DType) -> str:
        return "batched " if dtype.batched else ""

    @staticmethod
    def _ports_are_connected(p1: NodePort, p2: NodePort) -> bool:
        for c in p1.connections:
            if c in p2.connections:
                return True
        return False

    def check_connection_validity(self, p1: NodePort, p2: NodePort) -> bool:
        """Checks whether a considered connect action is legal"""

        # ryvencore.Flow.Flow content
        valid = True

        if p1.node == p2.node:
            valid = False

        if p1.io_pos == p2.io_pos or p1.type_ != p2.type_:
            valid = False

        # ironflow content
        if not self._ports_are_connected(p1, p2):
            # Only validate connections, not disconnections
            inp, out = (p1, p2) if p1.io_pos == 1 else (p2, p1)
            if (
                isinstance(inp.dtype, Untyped)
                or isinstance(out.dtype, Untyped)
                or (
                    inp.dtype.batched != out.dtype.batched
                    and isinstance(out.val, (list, np.ndarray))
                )
            ):
                type_valid = inp.dtype.accepts(out.val)
                check_type = "value"
            else:
                type_valid = inp.dtype.accepts(out.dtype)
                check_type = "dtype"
            InfoMsgs.write(
                f"{inp.node.title}.{inp.label_str} input "
                f"{self.batched_or_nothing(inp.dtype)}{inp.dtype.__class__.__name__} "
                f"made a {check_type} check to receive "
                f"{out.node.title}.{out.label_str} output "
                f"{self.batched_or_nothing(out.dtype)}{out.dtype.__class__.__name__} "
                f"and returned {type_valid}"
            )
            valid = valid and type_valid

        # ryvencore.Flow.Flow content
        self.connection_request_valid.emit(valid)

        return valid
